import re
import os
import filetype


def get_document_extension(document_path: str) -> str:
    """Get document extension

    :document_path: Path of the document
    :returns: Extension (string)

    """
    filetype.guess(document_path)
    kind = filetype.guess(document_path)
    if kind is None:
        m = re.match(r"^.*\.([^.]+)$", os.path.basename(document_path))
        return m.group(1) if m else 'data'
    else:
        assert isinstance(kind.extension, str)
        return kind.extension



"""

Cli
^^^
.. click:: papis.commands.rm:cli
    :prog: papis rm
"""
import click
import logging
import os

import papis.pick
import papis.tui.utils
import papis.document
import papis.cli
import papis.strings
import papis.database
import papis.git

from typing import Optional


def run(document: papis.document.Document,
        filepath: Optional[str] = None,
        git: bool = False) -> None:
    """Main method to the rm command
    """
    db = papis.database.get()
    _doc_folder = document.get_main_folder()
    if not _doc_folder:
        raise Exception(papis.strings.no_folder_attached_to_document)
    if filepath is not None:
        os.remove(filepath)
        document['files'].remove(os.path.basename(filepath))
        document.save()
        db.update(document)
        if git:
            papis.git.remove(_doc_folder, filepath)
            papis.git.add(_doc_folder, document.get_info_file())
            papis.git.commit(_doc_folder, "Remove file '{0}'".format(filepath))
    else:
        if git:
            _topfolder = os.path.dirname(os.path.abspath(_doc_folder))
            papis.git.remove(_doc_folder, _doc_folder, recursive=True)
            papis.git.commit(
                _topfolder,
                "Remove document '{0}'".format(
                    papis.document.describe(document)))
        else:
            papis.document.delete(document)
        db.delete(document)


@click.command("rm")
@click.help_option('-h', '--help')
@papis.cli.query_option()
@papis.cli.git_option(help="Remove in git")
@papis.cli.sort_option()
@papis.cli.doc_folder_option()
@click.option(
    "--file", "_file",
    help="Remove files from a document instead of the whole folder",
    is_flag=True,
    default=False)
@click.option(
    "-f", "--force",
    help="Do not confirm removal",
    is_flag=True,
    default=False)
@papis.cli.all_option()
def cli(query: str,
        git: bool,
        _file: bool,
        force: bool,
        _all: bool,
        doc_folder: str,
        sort_field: Optional[str],
        sort_reverse: bool) -> None:
    """Delete a document or a file"""

    if doc_folder:
        documents = [papis.document.from_folder(doc_folder)]
    else:
        documents = papis.database.get().query(query)

    if sort_field:
        documents = papis.document.sort(documents, sort_field, sort_reverse)

    logger = logging.getLogger('cli:rm')

    if not documents:
        logger.warning(papis.strings.no_documents_retrieved_message)
        return

    if not _all:
        documents = list(papis.pick.pick_doc(documents))

    if _file:
        for document in documents:
            filepaths = papis.pick.pick(document.get_files())
            if not filepaths:
                continue
            filepath = filepaths[0]
            if not force:
                tbar = 'The file {0} would be removed'.format(filepath)
                if not papis.tui.utils.confirm(
                        "Are you sure?", bottom_toolbar=tbar):
                    continue
            logger.info("Removing %s..." % filepath)
            run(document, filepath=filepath, git=git)
    else:
        for document in documents:
            if not force:
                tbar = 'The folder {0} would be removed'.format(
                    document.get_main_folder())
                logger.warning("This document will be removed, check it")
                papis.tui.utils.text_area(
                    title=tbar,
                    text=papis.document.dump(document),
                    lexer_name='yaml')
                if not papis.tui.utils.confirm(
                        "Are you sure?", bottom_toolbar=tbar):
                    continue
            logger.warning("removing ...")
            run(document, git=git)



"""This command is to update the information of the documents.

Some examples of the usage are given below

Examples
^^^^^^^^

- Update a document automatically and interactively
  (searching by ``doi`` number in *crossref*, or in other sources...)

    .. code::

        papis update --auto -i "author : dyson"

- Update your library from a bib(la)tex file where many entries are listed.
  papis will try to look for documents in your library that match these
  entries and will ask you entry per entry to update it (of course this is
  done if you use the ``-i`` flag for interactively doing it). In the example
  ``libraryfile.bib`` is a file containing many entries.

    .. code::

        papis update --from bibtex libraryfile.bib -i

- Tag all einstein papers with the tag classics

    .. code::

        papis update --all --set tags classics einstein

and add the tag of ``physics`` to all papers tagged as ``classics``

    .. code::

        papis update --all --set tags '{doc[tags]} physics' einstein

Cli
^^^
.. click:: papis.commands.update:cli
    :prog: papis update
"""

import click
import colorama
import logging

import papis.utils
import papis.tui.utils
import papis.strings
import papis.downloaders
import papis.document
import papis.database
import papis.pick
import papis.format
import papis.cli
import papis.importer
import papis.git

from typing import List, Dict, Tuple, Optional, Any


def _update_with_database(document: papis.document.Document) -> None:
    document.save()
    papis.database.get().update(document)


def run(document: papis.document.Document,
        data: Dict[str, Any] = dict(),
        git: bool = False) -> None:
    # Keep the ref the same, otherwise issues can be caused when
    # writing LaTeX documents and all the ref's change
    data['ref'] = document['ref']
    document.update(data)
    _update_with_database(document)
    folder = document.get_main_folder()
    info = document.get_info_file()
    if not folder or not info:
        raise Exception(papis.strings.no_folder_attached_to_document)
    if git:
        papis.git.add_and_commit_resource(
            folder, info,
            "Update information for '{0}'".format(
                papis.document.describe(document)))


@click.command("update")
@click.help_option('--help', '-h')
@papis.cli.git_option()
@papis.cli.query_option()
@papis.cli.doc_folder_option()
@papis.cli.all_option()
@papis.cli.sort_option()
@click.option("--auto",
              help="Try to parse information from different sources",
              default=False,
              is_flag=True)
@click.option("--from", "from_importer",
              help="Add document from a specific importer ({0})".format(
                ", ".join(papis.importer.available_importers())
                ),
              type=(click.Choice(papis.importer.available_importers()), str),
              nargs=2,
              multiple=True,
              default=(),)
@click.option("-s", "--set", "set_tuples",
              help="Update document's information with key value."
                   "The value can be a papis format.",
              multiple=True,
              type=(str, str),)
def cli(query: str,
        git: bool,
        doc_folder: str,
        from_importer: List[Tuple[str, str]],
        auto: bool,
        _all: bool,
        sort_field: Optional[str],
        sort_reverse: bool,
        set_tuples: List[Tuple[str, str]],) -> None:
    """Update a document from a given library."""
    logger = logging.getLogger('cli:update')

    if doc_folder:
        documents = [papis.document.from_folder(doc_folder)]
    else:
        documents = papis.database.get().query(query)

    if sort_field:
        documents = papis.document.sort(documents, sort_field, sort_reverse)

    if not _all:
        documents = list(papis.pick.pick_doc(documents))

    if not documents:
        logger.error(papis.strings.no_documents_retrieved_message)
        return

    for document in documents:
        ctx = papis.importer.Context()

        logger.info('Updating '
                    '{c.Back.WHITE}{c.Fore.BLACK}{0}{c.Style.RESET_ALL}'
                    .format(papis.document.describe(document), c=colorama))

        ctx.data.update(document)
        if set_tuples:
            ctx.data.update(
                {key: papis.format.format(value, document)
                    for key, value in set_tuples})

        matching_importers = []
        if not from_importer and auto:
            for importer_cls in papis.importer.get_importers():
                try:
                    importer = importer_cls.match_data(document)
                    if importer:
                        importer.fetch()
                except NotImplementedError:
                    continue
                except Exception as e:
                    logger.exception(e)
                else:
                    if importer and importer.ctx:
                        matching_importers.append(importer)

        for _importer_name, _uri in from_importer:
            try:
                _uri = papis.format.format(_uri, document)
                _iclass = papis.importer.get_importer_by_name(_importer_name)
                importer = _iclass(uri=_uri)
                importer.fetch()
                if importer.ctx:
                    matching_importers.append(importer)
            except Exception as e:
                logger.exception(e)

        if matching_importers:
            logger.info(
                'There are {0} possible matchings'
                .format(len(matching_importers)))

            for importer in matching_importers:
                if importer.ctx.data:
                    logger.info(
                        'Merging data from importer {0}'.format(importer.name))
                    papis.utils.update_doc_from_data_interactively(
                        ctx.data,
                        importer.ctx.data,
                        str(importer))
                if importer.ctx.files:
                    logger.info(
                        'Got files {0} from importer {1}'
                        .format(importer.ctx.files, importer.name))
                    for f in importer.ctx.files:
                        papis.utils.open_file(f)
                        if papis.tui.utils.confirm("Use this file?"):
                            ctx.files.append(f)

        run(document, data=ctx.data, git=git)



import logging
from typing import Optional, Union, Any, Dict

import papis.config
import papis.plugin
import papis.document
from papis.document import Document


FormatDocType = Union[Document, Dict[str, Any]]
LOGGER = logging.getLogger("format")
_FORMATER = None  # type: Optional[Formater]


class Formater:
    def format(self,
               fmt: str,
               doc: FormatDocType,
               key: str = "") -> str:
        """
        :param fmt: Python-like format string.
        :type  fmt: str
        :param doc: Papis document
        :type  doc: FormatDocType
        :returns: Formated string
        :rtype: str
        """
        ...


class PythonFormater(Formater):
    """Construct a string using a pythonic format string and a document.
    You can activate this formater by setting ``formater = python``.
    """
    def format(self,
               fmt: str,
               doc: FormatDocType,
               key: str = "") -> str:
        doc_name = key or papis.config.getstring("format-doc-name")
        fdoc = Document()
        fdoc.update(doc)
        try:
            return fmt.format(**{doc_name: fdoc})
        except Exception as exception:
            return str(exception)


class Jinja2Formater(Formater):
    """Construct a Jinja2 formated string.
    You can activate this formater by setting ``formater = jinja2``.
    """

    def __init__(self) -> None:
        try:
            import jinja2
        except ImportError as exception:
            LOGGER.exception("""
            You're trying to format strings using jinja2
            Jinja2 is not installed by default, so just install it
                pip3 install jinja2
            """)
            str(exception)
        else:
            self.jinja2 = jinja2

    def format(self,
               fmt: str,
               doc: FormatDocType,
               key: str = "") -> str:
        doc_name = key or papis.config.getstring("format-doc-name")
        try:
            return str(self.jinja2.Template(fmt).render(**{doc_name: doc}))
        except Exception as exception:
            return str(exception)


def _extension_name() -> str:
    return "papis.format"


def get_formater() -> Formater:
    """Get the formater named 'name' declared as a plugin"""
    global _FORMATER
    if _FORMATER is None:
        name = papis.config.getstring("formater")
        try:
            _FORMATER = papis.plugin.get_extension_manager(
                _extension_name())[name].plugin()
        except KeyError:
            LOGGER.error("Invalid formater (%s)", name)
            raise Exception(
                "Registered formaters are: %s",
                papis.plugin.get_available_entrypoints(_extension_name()))
        LOGGER.debug("Getting {}".format(name))
    return _FORMATER


def format(fmt: str,
           doc: FormatDocType,
           key: str = "") -> str:
    formater = get_formater()
    return formater.format(fmt, doc, key)



"""This module describes which functions are intended to be used by users to
create papis scripts.
"""
from typing import Any, Dict, List, Optional
import logging

import papis.utils
import papis.commands
import papis.config
import papis.pick
import papis.database

LOGGER = logging.getLogger("api")
LOGGER.debug("importing")


def get_lib_name() -> str:
    """Get current library, it either retrieves the library from
    the environment PAPIS_LIB variable or from the command line
    args passed by the user.

    :returns: Library name
    :rtype:  str

    >>> get_lib_name() is not None
    True
    """
    return papis.config.get_lib_name()


def set_lib_from_name(library: str) -> None:
    """Set current library, it either sets the library in
    the environment PAPIS_LIB variable or in the command line
    args passed by the user.

    :param library: Name of library or path to a given library
    :type  library: str

    """
    papis.config.set_lib_from_name(library)


def get_libraries() -> List[str]:
    """Get all libraries declared in the configuration. A library is discovered
    if the ``dir`` or ``dirs`` key defined in the library section.

    :returns: List of library names
    :rtype: list

    >>> len(get_libraries()) >= 1
    True

    """
    libs = []
    config = papis.config.get_configuration()
    for key in config:
        if "dir" in config[key] or "dirs" in config[key]:
            libs.append(key)
    return libs


pick_doc = papis.pick.pick_doc
pick = papis.pick.pick


def open_file(file_path: str, wait: bool = True) -> None:
    """Open file using the ``opentool`` key value as a program to
    handle file_path.

    :param file_path: File path to be handled.
    :type  file_path: str
    :param wait: Wait for the completion of the opener program to continue
    :type  wait: bool

    """
    papis.utils.open_file(file_path, wait=wait)


def open_dir(dir_path: str, wait: bool = True) -> None:
    """Open dir using the ``file-browser`` key value as a program to
    open dir_path.

    :param dir_path: Folder path to be handled.
    :type  dir_path: str
    :param wait: Wait for the completion of the opener program to continue
    :type  wait: bool

    """
    papis.utils.general_open(dir_path, "file-browser", wait=wait)


def edit_file(file_path: str, wait: bool = True) -> None:
    """Edit file using the ``editor`` key value as a program to
    handle file_path.

    :param file_path: File path to be handled.
    :type  file_path: str
    :param wait: Wait for the completion of the opener program to continue
    :type  wait: bool

    """
    papis.utils.general_open(file_path, "editor", wait=wait)


def get_all_documents_in_lib(
        library: Optional[str] = None) -> List[papis.document.Document]:
    """Get ALL documents contained in the given library with possibly.

    :param library: Library name.
    :type  library: str

    :returns: List of all documents.
    :rtype: list

    >>> import tempfile
    >>> folder = tempfile.mkdtemp()
    >>> set_lib_from_name(folder)
    >>> docs = get_all_documents_in_lib(folder)
    >>> len(docs)
    0

    """
    return papis.database.get(library).get_all_documents()


def get_documents_in_dir(
        directory: str, search: str = "") -> List[papis.document.Document]:
    """Get documents contained in the given folder with possibly a search
    string.

    :param directory: Folder path.
    :type  directory: str

    :param search: Search string
    :type  search: str

    :returns: List of filtered documents.
    :rtype: list

    >>> import tempfile
    >>> docs = get_documents_in_dir(tempfile.mkdtemp())
    >>> len(docs)
    0

    """
    set_lib_from_name(directory)
    return get_documents_in_lib(directory, search)


def get_documents_in_lib(
        library: Optional[str] = None,
        search: str = "") -> List[papis.document.Document]:
    """Get documents contained in the given library with possibly a search
    string.

    :param library: Library name.
    :type  library: str

    :param search: Search string
    :type  search: str

    :returns: List of filtered documents.
    :rtype: list

    """
    return papis.database.get(library).query(search)


def clear_lib_cache(lib: Optional[str] = None) -> None:
    """Clear cache associated with a library. If no library is given
    then the current library is used.

    :param lib: Library name.
    :type  lib: str

    >>> clear_lib_cache()

    """
    papis.database.get(lib).clear()


def doi_to_data(doi: str) -> Dict[str, Any]:
    """Try to get from a DOI expression a dictionary with the document's data
    using the crossref module.

    :param doi: DOI expression.
    :type  doi: str
    :returns: Document's data
    :rtype: dict
    """
    import papis.crossref
    return papis.crossref.doi_to_data(doi)



import re
import papis.downloaders.base

from typing import Optional


class Downloader(papis.downloaders.Downloader):

    def __init__(self, url: str):
        papis.downloaders.Downloader.__init__(
            self, url, name="frontiersin")
        self.expected_document_extension = 'pdf'
        self.cookies = {'gdpr': 'true'}

    @classmethod
    def match(cls, url: str) -> Optional[papis.downloaders.Downloader]:
        if re.match(r".*frontiersin.org.*", url):
            return Downloader(url)
        else:
            return None

    def get_doi(self) -> Optional[str]:
        url = self.uri
        self.logger.info('Parsing doi from {0}'.format(url))
        mdoi = re.match(r'.*/articles/([^/]+/[^/?&%^$]+).*', url)
        if mdoi:
            doi = mdoi.group(1)
            return doi
        return None

    def get_document_url(self) -> Optional[str]:
        durl = ("https://www.frontiersin.org/articles/{doi}/pdf"
                .format(doi=self.get_doi()))
        self.logger.debug("[doc url] = %s" % durl)
        return durl

    def get_bibtex_url(self) -> Optional[str]:
        url = ("https://www.frontiersin.org/articles/{doi}/bibTex"
               .format(doi=self.get_doi()))
        self.logger.debug("[bibtex url] = %s" % url)
        return url



import os
import glob
from typing import List


class Library:

    def __init__(self, name: str, paths: List[str]):
        """Create a Library object."""
        self.name = name
        self.paths = sum(
            [glob.glob(os.path.expanduser(p)) for p in paths],
            [])  # type: List[str]

    def path_format(self) -> str:
        return ":".join(self.paths)

    def __str__(self) -> str:
        """Get the name of the library when formatting"""
        return self.name


def from_paths(paths: List[str]) -> Library:
    name = ":".join(paths)
    return Library(name, paths)



import yaml
import logging
import click
import os
from typing import Optional, List, Dict, Any

import papis.utils
import papis.config
import papis.importer
import papis.document

logger = logging.getLogger("yaml")


def data_to_yaml(yaml_path: str, data: Dict[str, Any]) -> None:
    """
    Save data to yaml at path outpath

    :param yaml_path: Path to a yaml file
    :type  yaml_path: str
    :param data: Data in a dictionary
    :type  data: dict
    """
    with open(yaml_path, 'w+') as fd:
        yaml.dump(
            data,
            fd,
            allow_unicode=papis.config.getboolean("info-allow-unicode"),
            default_flow_style=False)


def exporter(documents: List[papis.document.Document]) -> str:
    string = yaml.dump_all(
        [papis.document.to_dict(document) for document in documents],
        allow_unicode=True)
    return str(string)


def yaml_to_data(
        yaml_path: str,
        raise_exception: bool = False) -> Dict[str, Any]:
    """
    Convert a yaml file into a dictionary using the yaml module.

    :param yaml_path: Path to a yaml file
    :type  yaml_path: str
    :returns: Dictionary containing the info of the yaml file
    :rtype:  dict
    :raises ValueError: If a yaml parsing error happens
    """
    global logger
    with open(yaml_path) as fd:
        try:
            data = yaml.safe_load(fd)
        except Exception as e:
            if raise_exception:
                raise ValueError(e)
            logger.error("Yaml syntax error: \n\n{0}".format(e))
            return dict()
        else:
            assert isinstance(data, dict)
            return data


@click.command('yaml')
@click.pass_context
@click.argument('yamlfile', type=click.Path(exists=True))
@click.help_option('--help', '-h')
def explorer(ctx: click.Context, yamlfile: str) -> None:
    """
    Import documents from a yaml file

    Examples of its usage are

    papis explore yaml lib.yaml pick

    """
    logger = logging.getLogger('explore:yaml')
    logger.info('reading in yaml file {}'.format(yamlfile))
    docs = [papis.document.from_data(d)
            for d in yaml.safe_load_all(open(yamlfile))]
    ctx.obj['documents'] += docs
    logger.info('{} documents found'.format(len(docs)))


class Importer(papis.importer.Importer):

    """Importer that parses a yaml file"""

    def __init__(self, uri: str) -> None:
        papis.importer.Importer.__init__(self, name='yaml', uri=uri)

    @classmethod
    def match(cls, uri: str) -> Optional[papis.importer.Importer]:
        importer = Importer(uri=uri)
        if os.path.exists(uri) and not os.path.isdir(uri):
            importer.fetch()
            return importer if importer.ctx.data else None
        return None

    @papis.importer.cache
    def fetch(self: papis.importer.Importer) -> Any:
        self.ctx.data = yaml_to_data(self.uri, raise_exception=False)
        if self.ctx:
            self.logger.info("successfully read file = %s" % self.uri)



import re
import papis.downloaders.fallback

from typing import Optional


class Downloader(papis.downloaders.fallback.Downloader):

    BASE = "https://citeseerx.ist.psu.edu"
    jsessionid = "012341666D7AD1C5C931FC0CFBA34BFA"

    def __init__(self, url: str):
        papis.downloaders.fallback.Downloader.__init__(self,
                                                       uri=url,
                                                       name="citeseerx")
        self.expected_document_extension = 'pdf'
        self.priority = 10

    @classmethod
    def match(cls,
              url: str) -> Optional[papis.downloaders.fallback.Downloader]:
        return (Downloader(url)
                if re.match(r".*citeseerx\.ist\.psu\.edu.*", url)
                else None)

    def get_document_url(self) -> Optional[str]:
        return ("{base}/viewdoc/download;jsessionid={jid}?"
                "doi={doi}&rep=rep1&type=pdf"
                .format(doi=self.ctx.data.get("doi"),
                        base=self.BASE,
                        jid=self.jsessionid))

    def download_bibtex(self) -> None:
        bibtex = self._get_soup().find_all(attrs=dict(id="bibtex"))
        if not bibtex:
            return None
        self.bibtex_data = ("\n".join(e.text for e in bibtex[0].find_all("p"))
                                .replace("\xa0", " "))



from typing import List, Optional, Any, Sequence, Type, Dict
import logging
import os
import re
import tempfile

import requests
import bs4
import filetype

import papis.bibtex
import papis.config
import papis.document
import papis.importer
import papis.plugin
import papis.utils

LOGGER = logging.getLogger("downloader")


def _extension_name() -> str:
    return "papis.downloader"


class Importer(papis.importer.Importer):

    """Importer that tries to get data and files from implemented downloaders
    """

    def __init__(self, **kwargs: Any):
        papis.importer.Importer.__init__(self, name='url', **kwargs)

    @classmethod
    def match(cls, uri: str) -> Optional[papis.importer.Importer]:
        return (
            Importer(uri=uri)
            if re.match(' *http(s)?.*', uri) is not None
            else None
        )

    def fetch(self) -> None:
        self.logger.info("attempting to import from url %s", self.uri)
        self.ctx = get_info_from_url(self.uri) or papis.importer.Context()

    def fetch_data(self) -> None:
        pass

    def fetch_files(self) -> None:
        pass


class Downloader(papis.importer.Importer):

    """This is the base class for every downloader.
    """

    def __init__(self,
                 uri: str = "",
                 name: str = "",
                 ctx: papis.importer.Context = papis.importer.Context()):
        papis.importer.Importer.__init__(self,
                                         uri=uri,
                                         ctx=ctx,
                                         name=name or
                                         os.path.basename(__file__))
        self.logger = logging.getLogger("downloader:"+self.name)
        self.logger.debug("uri {0}".format(uri))
        self.expected_document_extension = None  # type: Optional[str]
        self.priority = 1  # type: int
        self._soup = None  # type: Optional[bs4.BeautifulSoup]

        self.bibtex_data = None  # type: Optional[str]
        self.document_data = None  # type: Optional[bytes]

        self.session = requests.Session()  # type: requests.Session
        self.session.headers = requests.structures.CaseInsensitiveDict({
            'User-Agent': papis.config.getstring('user-agent')
        })
        proxy = papis.config.get('downloader-proxy')
        if proxy is not None:
            self.session.proxies = {
                'http': proxy,
                'https': proxy,
            }
        self.cookies = {}  # type: Dict[str, str]

    def fetch_data(self) -> None:
        """
        Try first to get data by hand with the get_data command.
        Then commplement with bibtex data.
        """
        # Try with get_data
        try:
            data = self.get_data()
            assert(isinstance(data, dict))
        except NotImplementedError:
            pass
        else:
            self.ctx.data.update(data)

        # try with bibtex
        try:
            self.download_bibtex()
        except NotImplementedError:
            pass
        else:
            bib_rawdata = self.get_bibtex_data()
            if bib_rawdata:
                datalist = papis.bibtex.bibtex_to_dict(bib_rawdata)
                if datalist:
                    self.logger.info("Merging data from bibtex")
                    self.ctx.data.update(datalist[0])
        # try getting doi
        try:
            doi = self.get_doi()
        except NotImplementedError:
            pass
        else:
            self.ctx.data['doi'] = doi

    def fetch_files(self) -> None:
        # get documents
        try:
            self.download_document()
        except NotImplementedError:
            pass
        else:
            doc_rawdata = self.get_document_data()
            if doc_rawdata and self.check_document_format():
                with tempfile.NamedTemporaryFile(
                        mode="wb+", delete=False) as f:
                    f.write(doc_rawdata)
                    self.logger.info("Saving downloaded file in %s", f.name)
                    self.ctx.files.append(f.name)

    def fetch(self) -> None:
        self.fetch_data()
        self.fetch_files()

    @classmethod
    def match(cls, url: str) -> Optional['Downloader']:
        raise NotImplementedError(
            "Matching uri not implemented for this importer")

    def _get_body(self) -> bytes:
        """Get body of the uri, this is also important for unittesting"""
        return self.session.get(self.uri).content

    def _get_soup(self) -> bs4.BeautifulSoup:
        """Get body of the uri, this is also important for unittesting"""
        if self._soup:
            return self._soup
        self._soup = bs4.BeautifulSoup(self._get_body(), features='lxml')
        return self._soup

    def __str__(self) -> str:
        return 'Downloader({0}, uri={1})'.format(self.name, self.uri)

    def get_bibtex_url(self) -> Optional[str]:
        """It returns the urls that is to be access to download
        the bibtex information. It has to be implemented for every
        downloader, or otherwise it will raise an exception.

        :returns: Bibtex url
        :rtype:  str
        """
        raise NotImplementedError(
            "Getting bibtex url not implemented for this downloader")

    def get_bibtex_data(self) -> Optional[str]:
        """Get the bibtex_data data if it has been downloaded already
        and if not download it and return the data in utf-8 format.

        :returns: Bibtex data in utf-8 format
        :rtype:  str
        """
        if not self.bibtex_data:
            self.download_bibtex()
        return self.bibtex_data

    def download_bibtex(self) -> None:
        """Bibtex downloader, it should try to download bibtex information from
        the url provided by ``get_bibtex_url``.

        It sets the ``bibtex_data`` attribute if it succeeds.

        :returns: Nothing
        :rtype:  None
        """
        url = self.get_bibtex_url()
        if not url:
            return
        res = self.session.get(url, cookies=self.cookies)
        self.logger.info("downloading bibtex from {0}".format(url))
        self.bibtex_data = res.content.decode()

    def get_document_url(self) -> Optional[str]:
        """It returns the urls that is to be access to download
        the document information. It has to be implemented for every
        downloader, or otherwise it will raise an exception.

        :returns: Document url
        :rtype:  str
        """
        raise NotImplementedError(
            "Getting bibtex url not implemented for this downloader")

    def get_data(self) -> Dict[str, Any]:
        """A general data retriever, for instance when data needn't need
        to come from a bibtex
        """
        raise NotImplementedError(
            "Getting general data is not implemented for this downloader")

    def get_doi(self) -> Optional[str]:
        """It returns the doi of the document, if it is retrievable.
        It has to be implemented for every downloader, or otherwise it will
        raise an exception.

        :returns: Document doi
        :rtype:  str
        """
        raise NotImplementedError(
            "Getting document url not implemented for this downloader")

    def get_document_data(self) -> Optional[bytes]:
        """Get the document_data data if it has been downloaded already
        and if not download it and return the data in binary format.

        :returns: Document data in binary format
        :rtype:  str
        """
        if not self.document_data:
            self.download_document()
        return self.document_data

    def download_document(self) -> None:
        """Document downloader, it should try to download document information from
        the url provided by ``get_document_url``.

        It sets the ``document_data`` attribute if it succeeds.

        :returns: Nothing
        :rtype:  None
        """
        url = self.get_document_url()
        if not url:
            return
        self.logger.info("downloading file from {0}".format(url))
        res = self.session.get(url, cookies=self.cookies)
        self.document_data = res.content

    def check_document_format(self) -> bool:
        """Check if the downloaded document has the filetype that the
        downloader expects. If the downloader does not expect any special
        filetype, accept anything because there is no way to know if it is
        correct.

        :returns: True if it is of the right type, else otherwise
        :rtype:  bool
        """
        def print_warning() -> None:
            self.logger.error(
                "The downloaded data does not seem to be of"
                "the correct type ({})"
                .format(self.expected_document_extension))

        if self.expected_document_extension is None:
            return True

        retrieved_kind = filetype.guess(self.get_document_data())

        if retrieved_kind is None:
            print_warning()
            return False

        self.logger.debug(
            'retrieved kind of document seems to be {0}'
            .format(retrieved_kind.mime))

        if not isinstance(self.expected_document_extension, list):
            expected_document_extensions = [self.expected_document_extension]

        if retrieved_kind.extension in expected_document_extensions:
            return True
        else:
            print_warning()
            return False


def get_available_downloaders() -> List[Type[Downloader]]:
    """Get all declared downloader classes
    """
    _ret = papis.plugin.get_available_plugins(
        _extension_name())  # type: List[Type[Downloader]]
    return _ret


def get_matching_downloaders(url: str) -> Sequence[Downloader]:
    """Get matching downloaders sorted by their priorities.
    The first elements have the higher priority

    :param url: Url to be matched against
    :type  url: str
    :returns: A list of sorted downloaders
    :rtype: list
    """
    print(get_available_downloaders())
    _maybe_matches = [
        d.match(url)
        for d in get_available_downloaders()]  # List[Optional[Downloader]]
    matches = [m
               for m in _maybe_matches
               if m is not None]  # type: List[Downloader]
    print(matches)
    return sorted(
        matches,
        key=lambda k: k.priority,
        reverse=True)


def get_downloader_by_name(name: str) -> Type[Downloader]:
    """Get a downloader by its name

    :param name: Name of the downloader
    :type  name: str
    :returns: A downloader class
    :rtype:  papis.Downloader
    """
    downloader_class = papis.plugin.get_extension_manager(
        _extension_name())[name].plugin  # type: Type[Downloader]
    return downloader_class


def get_info_from_url(
        url: str,
        expected_doc_format: Optional[str] = None
        ) -> papis.importer.Context:
    """Get information directly from url

    :param url: Url of the resource
    :type  url: str
    :param expected_doc_format: override the doc format of the document
    :type  expected_doc_format: str
    :returns: Context object
    :rtype:  papis.importer.Context or None
    """

    downloaders = get_matching_downloaders(url)
    if not downloaders:
        LOGGER.warning("No matching downloader found for (%s)", url)
        return papis.importer.Context()

    LOGGER.debug('Found %s matching downloaders', len(downloaders))
    downloader = downloaders[0]

    LOGGER.info("Using downloader '%s'", downloader)
    if downloader.expected_document_extension is None and \
            expected_doc_format is not None:
        downloader.expected_document_extension = expected_doc_format
    downloader.fetch()
    return downloader.ctx



import re
import papis.downloaders.base
from typing import Optional


class Downloader(papis.downloaders.Downloader):

    def __init__(self, url: str):
        papis.downloaders.Downloader.__init__(
            self, url, name="scitationaip")
        self.expected_document_extension = 'pdf'

    @classmethod
    def match(cls, url: str) -> Optional[papis.downloaders.Downloader]:
        # http://aip.scitation.org/doi/10.1063/1.4873138
        if re.match(r".*(aip|aapt)\.scitation\.org.*", url):
            return Downloader(url)
        else:
            return None

    def get_doi(self) -> Optional[str]:
        mdoi = re.match(r'.*/doi/(.*/[^?&%^$]*).*', self.uri)
        if mdoi:
            doi = mdoi.group(1).replace("abs/", "").replace("full/", "")
            return doi
        else:
            return None

    def get_document_url(self) -> Optional[str]:
        # http://aip.scitation.org/doi/pdf/10.1063/1.4873138
        durl = ("http://aip.scitation.org/doi/pdf/{doi}"
                .format(doi=self.get_doi()))
        self.logger.debug("[doc url] = %s" % durl)
        return durl

    def get_bibtex_url(self) -> Optional[str]:
        url = ("http://aip.scitation.org/action/downloadCitation"
               "?format=bibtex&cookieSet=1&doi={doi}"
               .format(doi=self.get_doi()))
        self.logger.debug("[bibtex url] = %s" % url)
        return url



import click
import click.core
import click.types
import click.decorators

import papis.config
import difflib
from typing import Optional, Any, Callable


DecoratorCallable = Callable[..., Any]
DecoratorArgs = Any


class AliasedGroup(click.core.Group):
    """
    This group command is taken from
        http://click.palletsprojects.com/en/5.x/advanced/#command-aliases
    and is to be used for groups with aliases
    """

    def get_command(
            self,
            ctx: click.core.Context,
            cmd_name: str) -> Optional[click.core.Command]:
        rv = click.core.Group.get_command(self, ctx, cmd_name)
        if rv is not None:
            return rv
        matches = difflib.get_close_matches(
            cmd_name, self.list_commands(ctx), n=2)
        if not matches:
            return None
        elif len(matches) == 1:
            return click.core.Group.get_command(self, ctx, str(matches[0]))
        else:
            ctx.fail('Too many matches: {0}'.format(matches))
            return None


def query_option(**attrs: DecoratorArgs) -> DecoratorCallable:
    """Adds a ``query`` argument as a decorator"""
    def decorator(f: DecoratorCallable) -> Any:
        attrs.setdefault(
            'default',
            lambda: papis.config.get('default-query-string'))
        return click.decorators.argument('query', **attrs)(f)
    return decorator


def sort_option(**attrs: DecoratorArgs) -> DecoratorCallable:
    """Adds a ``sort`` argument as a decorator"""
    def decorator(f: DecoratorCallable) -> Any:
        attrs.setdefault('default', lambda: papis.config.get('sort-field'))
        attrs.setdefault('help', 'Sort documents with respect to FIELD')
        attrs.setdefault('metavar', 'FIELD')
        sort_f = click.decorators.option('--sort', "sort_field", **attrs)
        reverse_f = click.decorators.option(
            "--reverse", "sort_reverse",
            help="Reverse sort order", is_flag=True)
        return sort_f(reverse_f(f))
    return decorator


def doc_folder_option(**attrs: DecoratorArgs) -> DecoratorCallable:
    """Adds a ``document folder`` argument as a decorator"""
    def decorator(f: DecoratorCallable) -> Any:
        attrs.setdefault('default', None)
        attrs.setdefault('type', click.types.Path(exists=True))
        attrs.setdefault('help', 'Apply action to a document path')
        return click.decorators.option('--doc-folder', **attrs)(f)
    return decorator


def all_option(**attrs: DecoratorArgs) -> DecoratorCallable:
    """Adds a ``query`` argument as a decorator"""
    def decorator(f: DecoratorCallable) -> Any:
        attrs.setdefault('default', False)
        attrs.setdefault('is_flag', True)
        attrs.setdefault('help', 'Apply action to all matching documents')
        return click.decorators.option('-a', '--all', '_all', **attrs)(f)
    return decorator


def git_option(
        help: str = "Add git interoperability",
        **attrs: DecoratorArgs) -> DecoratorCallable:
    """Adds a ``git`` option as a decorator"""
    def decorator(f: DecoratorCallable) -> Any:
        attrs.setdefault(
            'default',
            lambda: True if papis.config.get('use-git') else False)
        attrs.setdefault('help', help)
        return click.decorators.option('--git/--no-git', **attrs)(f)
    return decorator


def bypass(
        group: click.core.Group,
        command: click.core.Command,
        command_name: str) -> Callable[..., Any]:
    """
    This function is specially important for people developing scripts in
    papis.

    Suppose you're writing a plugin that uses the ``add`` command as seen
    in the command line in papis. However you don't want exactly the ``add``
    command and you want to add some behavior before calling it, and you
    don't want to write your own ``add`` function from scratch.

    You can then use the following snippet

    .. code::python

        import click
        import papis.cli
        import papis.commands.add

        @click.group()
        def main():
            \"\"\"Your main app\"\"\"
            pass

        @papis.cli.bypass(main, papis.commands.add.cli, "add")
        def add(**kwargs):
            # do some logic here...
            # and call the original add command line function by
            papis.commands.add.cli.bypassed(**kwargs)
    """
    group.add_command(command, command_name)

    def _decorator(new_callback: Callable[..., Any]) -> None:
        setattr(command, "bypassed", command.callback)
        command.callback = new_callback
    return _decorator



"""
TODO: Create an own python package for this

For description refer to
https://www.base-search.net/about/download/base_interface.pdf

"""
import urllib.parse
import urllib.request  # import urlencode
import logging
import json
from typing import Optional, Dict, Any, List, Callable, NamedTuple

import click

LOGGER = logging.getLogger('base')


def get_data(query: str = "", max: int = 20) -> List[Dict[str, Any]]:
    base_baseurl = (
        "https://api.base-search.net/"
        "cgi-bin/BaseHttpSearchInterface.fcgi/"
    )

    LOGGER.warning('BASE engine in papis is experimental')

    if max > 125:
        LOGGER.error('BASE only allows a maximum of 125 hits')
        max = 125

    dict_params = {
        "func": "PerformSearch",
        "query": query,
        "format": "json",
        "hits": max,
    }
    params = urllib.parse.urlencode(
        {x: dict_params[x] for x in dict_params if dict_params[x]}
    )
    req_url = base_baseurl + "search?" + params
    LOGGER.debug("url = " + req_url)
    url = urllib.request.Request(
        req_url,
        headers={
            'User-Agent': 'papis'
        }
    )
    jsondoc = json.loads(urllib.request.urlopen(url).read().decode())
    docs = jsondoc.get('response').get('docs')
    LOGGER.info("Retrieved {0} documents".format(len(docs)))
    return list(map(basedoc_to_papisdoc, docs))


def basedoc_to_papisdoc(basedoc: Dict[str, Any]) -> Dict[str, Any]:
    """Convert a json doc from the base database into a papis document

    :basedoc: Json doc from base database
    :returns: Dictionary containing its data

    """
    doc = dict()
    _action_type = Optional[Callable[[List[str]], str]]
    _key_translate = NamedTuple("_key_translate", [("basekey", str),
                                                   ("papiskey", str),
                                                   ("mode", str),
                                                   ("action", _action_type)
                                                   ])
    keys_translate = [
        _key_translate("dctitle", "title", "s", None),
        _key_translate("dcyear", "year", "s", None),
        _key_translate("dclink", "url", "s", None),
        _key_translate("dcdescription", "abstract", "s", None),
        _key_translate("dcpublisher", "publisher", "m", lambda x: x[0]),
        _key_translate("dcperson", "author", "m", lambda x: " and ".join(x)),
        _key_translate("dcsubject", "tags", "m", lambda x: " ".join(x)),
        _key_translate("dcdoi", "doi", "m", lambda x: x[0]),
        _key_translate("dctype", "type", "m", lambda x: x[0].lower()),
        _key_translate("dclang", "lang", "mutli", lambda x: x[0]),
    ]  # type: List[_key_translate]
    for kt in keys_translate:
        if kt.basekey not in basedoc.keys():
            continue
        key = kt.papiskey
        if kt.mode == "m":
            value = basedoc[kt.basekey]
            value = kt.action(value) if kt.action is not None else value
        else:
            value = basedoc[kt.basekey]
        doc[key] = value
    return doc


@click.command('base')
@click.pass_context
@click.help_option('--help', '-h')
@click.option('--query', '-q', default=None)
def explorer(ctx: click.core.Context, query: str) -> None:
    """
    Look for documents on the BielefeldAcademicSearchEngine

    Examples of its usage are

    papis explore base -q 'Albert einstein' pick cmd 'firefox {doc[url]}'

    """
    import papis.document
    logger = logging.getLogger('explore:base')
    logger.info('Looking up...')
    data = get_data(query=query)
    docs = [papis.document.from_data(data=d) for d in data]
    ctx.obj['documents'] += docs
    logger.info('{} documents found'.format(len(docs)))



"""The following table lists the field prefixes for all the fields
 that can be searched.

 Table:          search_query field prefixes
============================================
 prefix          explanation
--------------------------------------------
 ti              Title
 au              Author
 abs             Abstract
 co              Comment
 jr              Journal Reference
 cat             Subject Category
 rn              Report Number
 id              Id (use id_list instead)
 all             All of the above
"""
import os
import re
import bs4
import logging
import urllib.request  # urlopen, Request
import urllib.parse  # import urlencode
from typing import Optional, List, Dict, Any

import click
import arxiv2bib

import papis.filetype
import papis.downloaders.base
import papis.config


logger = logging.getLogger('arxiv')


def get_data(
    query: str = "",
    author: str = "",
    title: str = "",
    abstract: str = "",
    comment: str = "",
    journal: str = "",
    report_number: str = "",
    category: str = "",
    id_list: str = "",
    page: int = 0,
    max_results: int = 30
        ) -> List[Dict[str, Any]]:
    dict_params = {
        "all": query,
        "ti": title,
        "au": author,
        "cat": category,
        "abs": abstract,
        "co": comment,
        "jr": journal,
        "id_list": id_list,
        "rn": report_number
    }
    result = []
    clean_params = {x: dict_params[x] for x in dict_params if dict_params[x]}
    search_query = '+AND+'.join(
        [key+':'+str(clean_params[key]) for key in clean_params]
    )
    logger.debug("query = " + search_query)
    params = urllib.parse.urlencode(
        {
            'search_query': search_query,
            'start': page,
            'max_results': max_results
        }
    )
    main_url = "http://arxiv.org/api/query?"
    req_url = main_url + params
    logger.debug("url = " + req_url)
    url = urllib.request.Request(
        req_url,
        headers={
            'User-Agent': papis.config.getstring('user-agent')
        }
    )
    xmldoc = urllib.request.urlopen(url).read()
    soup = bs4.BeautifulSoup(xmldoc, 'html.parser')

    entries = soup.find_all("entry")
    for entry in entries:
        data = dict()
        data["abstract"] = entry.find("summary").get_text().replace(
            "\n", " "
        )
        data["url"] = entry.find("id").get_text()
        data["published"] = entry.find("published").get_text()
        published = data.get("published")
        if published:
            assert(isinstance(published, str))
            data["year"] = published[0:4]
        data["title"] = entry.find("title").get_text().replace("\n", " ")
        data["author"] = ", ".join(
            [
                author.get_text().replace("\n", "")
                for author in entry.find_all("author")
            ]
        )
        result.append(data)
    return result


def validate_arxivid(arxivid: str) -> None:
    from urllib.error import HTTPError, URLError
    url = "https://arxiv.org/abs/{0}".format(arxivid)
    request = urllib.request.Request(url)

    try:
        urllib.request.urlopen(request)
    except HTTPError:
        raise ValueError('HTTP 404: {0} not an arxivid'.format(arxivid))
    except URLError:
        pass


def pdf_to_arxivid(
        filepath: str, maxlines: float = float('inf')) -> Optional[str]:
    """Try to get arxivid from a filepath, it looks for a regex in the binary
    data and returns the first arxivid found, in the hopes that this arxivid
    is the correct one.

    :param filepath: Path to the pdf file
    :type  filepath: str
    :param maxlines: Maximum number of lines that should be checked
        For some documnets, it would spend a long time trying to look for
        a arxivid, and arxivids in the middle of documents don't tend to be the
        correct arxivid of the document.
    :type  maxlines: int
    :returns: arxivid or None
    :rtype:  str or None
    """
    with open(filepath, 'rb') as fd:
        for j, line in enumerate(fd):
            arxivid = find_arxivid_in_text(
                line.decode('ascii', errors='ignore'))
            if arxivid:
                return arxivid
            if j > maxlines:
                return None
    return None


def find_arxivid_in_text(text: str) -> Optional[str]:
    """
    Try to find a arxivid in a text
    """
    forbidden_arxivid_characters = r'"\(\)\s%!$^\'<>@,;:#?&'
    # Sometimes it is in the javascript defined
    regex = re.compile(
        r'arxiv(.org|.com)?'
        r'(/abs|/pdf)?'
        r'\s*(=|:|/|\()\s*'
        r'("|\')?'
        r'(?P<arxivid>[^{fc}]+)'
        r'("|\'|\))?'
        .format(
            fc=forbidden_arxivid_characters
        ), re.I
    )
    miter = regex.finditer(text)
    try:
        m = next(miter)
        if m:
            arxivid = m.group('arxivid')
            return arxivid
    except StopIteration:
        pass
    return None


@click.command('arxiv')
@click.pass_context
@click.help_option('--help', '-h')
@click.option('--query', '-q', default="", type=str)
@click.option('--author', '-a', default="", type=str)
@click.option('--title', '-t', default="", type=str)
@click.option('--abstract', default="", type=str)
@click.option('--comment', default="", type=str)
@click.option('--journal', default="", type=str)
@click.option('--report-number', default="", type=str)
@click.option('--category', default="", type=str)
@click.option('--id-list', default="", type=str)
@click.option('--page', default=0, type=int)
@click.option('--max', '-m', default=20, type=int)
def explorer(
        ctx: click.core.Context,
        query: str, author: str, title: str, abstract: str, comment: str,
        journal: str, report_number: str, category: str, id_list: str,
        page: int, max: int) -> None:
    """
    Look for documents on ArXiV.org.

    Examples of its usage are

        papis explore arxiv -a 'Hummel' -m 100 arxiv -a 'Garnet Chan' pick

    If you want to search for the exact author name 'John Smith', you should
    enclose it in extra quotes, as in the example below

        papis explore arxiv -a '"John Smith"' pick

    """
    logger = logging.getLogger('explore:arxiv')
    logger.info('Looking up...')
    data = get_data(
        query=query,
        author=author,
        title=title,
        abstract=abstract,
        comment=comment,
        journal=journal,
        report_number=report_number,
        category=category,
        id_list=id_list,
        page=page or 0,
        max_results=max)
    docs = [papis.document.from_data(data=d) for d in data]
    ctx.obj['documents'] += docs
    logger.info('%s documents found', len(docs))


class Downloader(papis.downloaders.Downloader):

    def __init__(self, url: str):
        papis.downloaders.Downloader.__init__(self, uri=url, name="arxiv")
        self.expected_document_extension = 'pdf'
        self.arxivid = None  # type: Optional[str]

    @classmethod
    def match(cls, url: str) -> Optional[papis.downloaders.Downloader]:
        arxivid = find_arxivid_in_text(url)
        if arxivid:
            url = "https://arxiv.org/abs/{0}".format(arxivid)
            down = Downloader(url)
            down.arxivid = arxivid
            return down
        else:
            return None

    def _get_identifier(self) -> Optional[str]:
        """Get arxiv identifier
        :returns: Identifier
        """
        if not self.arxivid:
            self.arxivid = find_arxivid_in_text(self.uri)
        return self.arxivid

    def get_bibtex_url(self) -> Optional[str]:
        return self._get_identifier()

    def download_bibtex(self) -> None:
        bib_url = self.get_bibtex_url()
        if not bib_url:
            return None
        bibtex_cli = arxiv2bib.Cli([bib_url])
        bibtex_cli.run()
        self.logger.debug("[bibtex url] = %s" % bib_url)
        output = bibtex_cli.output  # List[str]
        data = ''.join(output).replace('\n', ' ')
        self.bibtex_data = data

    def get_document_url(self) -> Optional[str]:
        arxivid = self._get_identifier()
        self.logger.debug("arxivid %s" % arxivid)
        pdf_url = "https://arxiv.org/pdf/{arxivid}.pdf".format(arxivid=arxivid)
        self.logger.debug("[pdf url] = %s" % pdf_url)
        return pdf_url


class Importer(papis.importer.Importer):

    """Importer accepting an arxiv id and downloading files and data"""

    def __init__(self, uri: str):
        papis.importer.Importer.__init__(self, name='arxiv', uri=uri)
        self.downloader = Downloader('https://arxiv.org/abs/{0}'.format(uri))

    @classmethod
    def match(cls, uri: str) -> Optional[papis.importer.Importer]:
        try:
            validate_arxivid(uri)
        except ValueError:
            return None
        else:
            return Importer(uri=uri)

    def fetch(self) -> None:
        self.downloader.fetch()
        self.ctx = self.downloader.ctx


class ArxividFromPdfImporter(papis.importer.Importer):

    """Importer parsing an arxivid from a pdf file"""

    def __init__(self, uri: str):
        papis.importer.Importer.__init__(self, name='pdf2arxivid', uri=uri)
        self.arxivid = None  # type: Optional[str]

    @classmethod
    def match(cls, uri: str) -> Optional[papis.importer.Importer]:
        if (os.path.isdir(uri) or not os.path.exists(uri) or
                not papis.filetype.get_document_extension(uri) == 'pdf'):
            return None
        importer = ArxividFromPdfImporter(uri=uri)
        importer.arxivid = pdf_to_arxivid(uri, maxlines=2000)
        return importer if importer.arxivid else None

    def fetch(self) -> None:
        self.logger.info(
            "trying to parse arxivid from file {0}".format(self.uri))
        if not self.arxivid:
            self.arxivid = pdf_to_arxivid(self.uri, maxlines=2000)
        if self.arxivid:
            self.logger.info("Parsed arxivid {0}".format(self.arxivid))
            self.logger.warning(
                "There is no guarantee that this arxivid is the one")
            importer = Importer.match(self.arxivid)
            if importer:
                importer.fetch()
                self.ctx = importer.ctx



import pickle
import logging
import os
import papis.utils
import papis.docmatcher
import papis.document
import papis.config
import papis.format
import papis.database.base
import re
import multiprocessing
import time
import sys
from typing import List, Optional, Match, Dict, Tuple


logger = logging.getLogger("cache")


def get_cache_file_name(directory: str) -> str:
    """Create a cache file name out of the path of a given directory.

    :param directory: Folder name to be used as a seed for the cache name.
    :type  directory: str
    :returns: Name for the cache file.
    :rtype:  str

    >>> get_cache_file_name('path/to/my/lib')
    'a8c689820a94babec20c5d6269c7d488-lib'
    >>> get_cache_file_name('papers')
    'a566b2bebc62611dff4cdaceac1a7bbd-papers'
    """
    import hashlib
    return "{}-{}".format(
        hashlib.md5(directory.encode()).hexdigest(),
        os.path.basename(directory))


def get_cache_file_path(directory: str) -> str:
    """Get the full path to the cache file

    :param directory: Library folder
    :type  directory: str

    >>> import os; os.environ["XDG_CACHE_HOME"] = '/tmp'
    >>> os.path.basename(get_cache_file_path('blah/papers'))
    'c39177eca0eaea2e21134b0bd06631b6-papers'
    """
    cache_name = get_cache_file_name(directory)
    folder = os.path.expanduser(
        os.path.join(papis.utils.get_cache_home(), 'database'))
    if not os.path.exists(folder):
        os.makedirs(folder)
    return os.path.join(folder, cache_name)


def filter_documents(
        documents: List[papis.document.Document],
        search: str = "") -> List[papis.document.Document]:
    """Filter documents. It can be done in a multi core way.

    :param documents: List of papis documents.
    :type  documents: papis.documents.Document
    :param search: Valid papis search string.
    :type  search: str
    :returns: List of filtered documents
    :rtype:  list

    >>> document = papis.document.from_data({'author': 'einstein'})
    >>> len(filter_documents([document], search="einstein")) == 1
    True
    >>> len(filter_documents([document], search="author : ein")) == 1
    True
    >>> len(filter_documents([document], search="title : ein")) == 1
    False

    """
    logger = logging.getLogger('filter')
    papis.docmatcher.DocMatcher.set_search(search)
    papis.docmatcher.DocMatcher.parse()
    papis.docmatcher.DocMatcher.set_matcher(match_document)
    begin_t = 1000 * time.time()
    # FIXME: find a better solution for this that works for both OSes
    if sys.platform == "win32":
        logger.debug(
            "Filtering {0} docs (search {1})".format(
                len(documents), search))
        filtered_docs = [
            d for d in [papis.docmatcher.DocMatcher.return_if_match(d)
                        for d in documents] if d is not None]

    else:
        # Doing this multiprocessing in filtering does not seem
        # to help much, I don't know if it's because I'm doing something
        # wrong or it is really like this.
        np = multiprocessing.cpu_count()
        pool = multiprocessing.Pool(np)
        logger.debug(
            "Filtering {0} docs (search {1}) using {2} cores".format(
                len(documents), search, np))
        result = pool.map(papis.docmatcher.DocMatcher.return_if_match,
                          documents)
        pool.close()
        pool.join()
        filtered_docs = [d for d in result if d is not None]
    _delta = 1000 * time.time() - begin_t
    logger.debug("done ({0} ms) ({1} docs)".format(_delta, len(filtered_docs)))
    return filtered_docs


def match_document(
        document: papis.document.Document, search: str,
        match_format: Optional[str] = None) -> Optional[Match[str]]:
    """Main function to match document to a given search.

    :param document: Papis document
    :type  document: papis.document.Document
    :param search: A valid search string
    :type  search: str
    :param match_format: Python-like format string.
        (`see <
            https://docs.python.org/2/library/string.html#format-string-syntax
        >`_)
    :type  match_format: str
    :returns: Non false if matches, true-ish if it does match.

    >>> papis.config.set('match-format', '{doc[author]}')
    >>> document = papis.document.from_data({'author': 'einstein'})
    >>> match_document(document, 'e in') is None
    False
    >>> match_document(document, 'ee in') is None
    True
    >>> match_document(document, 'einstein', '{doc[title]}') is None
    True
    """
    match_format = match_format or str(papis.config.get("match-format"))
    match_string = papis.format.format(match_format, document)
    regex = get_regex_from_search(search)
    return re.match(regex, match_string, re.IGNORECASE)


def get_regex_from_search(search: str) -> str:
    r"""Creates a default regex from a search string.

    :param search: A valid search string
    :type  search: str
    :returns: Regular expression
    :rtype: str

    >>> get_regex_from_search(' ein 192     photon')
    '.*ein.*192.*photon.*'

    >>> get_regex_from_search('{1234}')
    '.*\\{1234\\}.*'
    """
    return ".*" + ".*".join(map(re.escape, search.split())) + ".*"


class Database(papis.database.base.Database):

    def __init__(self, library: Optional[papis.library.Library] = None):
        papis.database.base.Database.__init__(self, library)
        self.logger = logging.getLogger('db:cache')
        self.documents = None  # type: Optional[List[papis.document.Document]]
        self.initialize()

    def get_backend_name(self) -> str:
        return 'papis'

    def initialize(self) -> None:
        pass

    def get_documents(self) -> List[papis.document.Document]:
        if self.documents is not None:
            return self.documents
        use_cache = papis.config.getboolean("use-cache")
        cache_path = self._get_cache_file_path()
        if use_cache and os.path.exists(cache_path):
            self.logger.debug(
                "Getting documents from cache in {0}".format(cache_path))
            with open(cache_path, 'rb') as fd:
                self.documents = pickle.load(fd)
        else:
            self.logger.info('Indexing library, this might take a while')
            folders = sum([
                papis.utils.get_folders(d)
                for d in self.get_dirs()], [])  # type: List[str]
            self.documents = papis.utils.folders_to_documents(folders)
            if use_cache:
                self.save()
        self.logger.debug(
            "Loaded documents (%s documents)", len(self.documents))
        return self.documents

    def add(self, document: papis.document.Document) -> None:
        docs = self.get_documents()
        self.logger.debug('adding ...')
        docs.append(document)
        assert(docs[-1].get_main_folder() == document.get_main_folder())
        _folder = document.get_main_folder()
        assert(_folder is not None)
        assert(os.path.exists(_folder))
        self.save()

    def update(self, document: papis.document.Document) -> None:
        if not papis.config.getboolean("use-cache"):
            return
        docs = self.get_documents()
        self.logger.debug('updating document')
        result = self._locate_document(document)
        index = result[0][0]
        docs[index] = document
        self.save()

    def delete(self, document: papis.document.Document) -> None:
        if not papis.config.getboolean("use-cache"):
            return
        docs = self.get_documents()
        self.logger.debug('deleting document')
        result = self._locate_document(document)
        index = result[0][0]
        docs.pop(index)
        self.save()

    def match(
            self, document: papis.document.Document,
            query_string: str) -> bool:
        return bool(match_document(document, query_string))

    def clear(self) -> None:
        cache_path = self._get_cache_file_path()
        self.logger.warning("clearing cache {0}".format(cache_path))
        if os.path.exists(cache_path):
            os.remove(cache_path)

    def query_dict(
            self, dictionary: Dict[str, str]) -> List[papis.document.Document]:
        query_string = " ".join(
            ["{}:\"{}\" ".format(key, val)
                for key, val in dictionary.items()])
        return self.query(query_string)

    def query(self, query_string: str) -> List[papis.document.Document]:
        self.logger.debug('Querying')
        docs = self.get_documents()
        # This makes it faster, if it's the all query string, return everything
        # without filtering
        if query_string == self.get_all_query_string():
            return docs
        else:
            return filter_documents(docs, query_string)

    def get_all_query_string(self) -> str:
        return '.'

    def get_all_documents(self) -> List[papis.document.Document]:
        return self.get_documents()

    def save(self) -> None:
        docs = self.get_documents()
        self.logger.debug(
            'Saving ... ({} documents)'.format(len(docs)))
        path = self._get_cache_file_path()
        with open(path, "wb+") as fd:
            pickle.dump(docs, fd)

    def _get_cache_file_path(self) -> str:
        return get_cache_file_path(self.lib.path_format())

    def _locate_document(
            self,
            document: papis.document.Document
            ) -> List[Tuple[int, papis.document.Document]]:
        assert(isinstance(document, papis.document.Document))
        result = list(filter(
            lambda d: d[1].get_main_folder() == document.get_main_folder(),
            enumerate(self.get_documents())))
        if len(result) == 0:
            raise Exception(
                'The document passed could not be found in the library')
        return result



import re
import papis.downloaders.base
import papis.document
from typing import Optional, Any, Dict


class Downloader(papis.downloaders.Downloader):

    def __init__(self, url: str):
        papis.downloaders.Downloader.__init__(
            self, url, name="springer")
        self.expected_document_extension = 'pdf'
        self.priority = 10

    @classmethod
    def match(cls, url: str) -> Optional[papis.downloaders.Downloader]:
        return (Downloader(url)
                if re.match(r".*link\.springer.com.*", url) else None)

    def get_data(self) -> Dict[str, Any]:
        data = dict()
        soup = self._get_soup()
        metas = soup.find_all(name="meta")
        author_list = []
        for meta in metas:
            if meta.attrs.get('name') == 'citation_title':
                data['title'] = meta.attrs.get('content')
            elif meta.attrs.get('name') == 'dc.Subject':
                data['subject'] = meta.attrs.get('content')
            elif meta.attrs.get('name') == 'citation_doi':
                data['doi'] = meta.attrs.get('content')
            elif meta.attrs.get('name') == 'citation_publisher':
                data['publisher'] = meta.attrs.get('content')
            elif meta.attrs.get('name') == 'citation_journal_title':
                data['journal'] = meta.attrs.get('content')
            elif meta.attrs.get('name') == 'citation_issn':
                data['issn'] = meta.attrs.get('content')
            elif meta.attrs.get('name') == 'citation_author':
                fnam = meta.attrs.get('content')
                fnams = re.split(r'\s+', fnam)
                author_list.append(
                    dict(
                        given=fnams[0],
                        family=' '.join(fnams[1:]),
                        affiliation=[]))
            elif meta.attrs.get('name') == 'citation_author_institution':
                if not author_list:
                    continue
                author_list[-1]['affiliation'].append(
                    dict(name=meta.attrs.get('content')))

        data['author_list'] = author_list
        data['author'] = papis.document.author_list_to_author(data)

        return data

    def get_bibtex_url(self) -> Optional[str]:
        if 'doi' in self.ctx.data:
            url = ("http://citation-needed.springer.com/v2/"
                   "references/{doi}?format=bibtex&amp;flavour=citation"
                   .format(doi=self.ctx.data['doi']))
            self.logger.debug("bibtex url = %s" % url)
            return url
        else:
            return None

    def get_document_url(self) -> Optional[str]:
        if 'doi' in self.ctx.data:
            url = ("https://link.springer.com/content/pdf/"
                   "{doi}.pdf".format(doi=self.ctx.data['doi']))
            self.logger.debug("doc url = %s" % url)
            return url
        else:
            return None



import re
import papis.downloaders.fallback
from typing import Optional


class Downloader(papis.downloaders.fallback.Downloader):

    def __init__(self, url: str):
        papis.downloaders.fallback.Downloader.__init__(self, url, name="hal")
        self.expected_document_extension = 'pdf'
        self.priority = 10

    @classmethod
    def match(
            cls, url: str) -> Optional[papis.downloaders.fallback.Downloader]:
        if re.match(r".*hal\.archives-ouvertes\.fr.*", url):
            return Downloader(url)
        else:
            return None

    def get_bibtex_url(self) -> Optional[str]:
        if 'pdf_url' in self.ctx.data:
            url = re.sub(r'document', 'bibtex', self.uri)
            self.logger.debug('bibtex url = {url}'.format(url=url))
            return url
        else:
            return None



import logging
from typing import Optional, List, Dict, Any, Callable, Type
import os.path

import papis
import papis.plugin


class Context:
    def __init__(self) -> None:
        self.data = dict()  # type: Dict[str, Any]
        self.files = []  # type: List[str]

    def __bool__(self) -> bool:
        return bool(self.files) or bool(self.data)


class Importer:

    """This is the base class for every importer"""

    def __init__(self, uri: str = "", name: str = "",
                 ctx: Optional[Context] = None):
        """
        :param uri: uri
        :type  uri: str
        :param name: Name of the importer
        :type  name: str
        """
        self.ctx = ctx or Context()  # type: Context
        self.uri = uri  # type: str
        self.name = name or os.path.basename(__file__)  # type: str
        self.logger = logging.getLogger("importer:{0}".format(self.name))

    @classmethod
    def match(cls, uri: str) -> Optional['Importer']:
        """This method should be called to know if a given uri matches
        the importer or not.

        For example, a valid match for archive would be:
        .. code:: python

            return re.match(r".*arxiv.org.*", uri)

        it will return something that is true if it matches and something
        falsely otherwise.

        :param uri: uri where the document should be retrieved from.
        :type  uri: str
        """
        raise NotImplementedError(
            "Matching uri not implemented for this importer")

    @classmethod
    def match_data(cls, data: Dict[str, Any]) -> Optional['Importer']:
        """Get a dictionary of data and try to decide if there is
        a valid uri in it.

        :param data: Data to look into
        :type  data: dict
        """
        raise NotImplementedError(
            "Matching data not implemented for this importer")

    def fetch(self) -> None:
        """
        can return a dict to update the document with
        """
        try:
            self.fetch_data()
        except NotImplementedError:
            pass
        try:
            self.fetch_files()
        except NotImplementedError:
            pass

    def fetch_data(self) -> None:
        raise NotImplementedError()

    def fetch_files(self) -> None:
        raise NotImplementedError()

    def __str__(self) -> str:
        return 'Importer({0}, uri={1})'.format(self.name, self.uri)


def _extension_name() -> str:
    return "papis.importer"


def get_import_mgr() -> papis.plugin.ExtensionManager:
    """Get the import manager
    :returns: Import manager
    """
    return papis.plugin.get_extension_manager(_extension_name())


def available_importers() -> List[str]:
    """Get the available importer names.
    :returns: List of importer names
    :rtype:  list(str)
    """
    return papis.plugin.get_available_entrypoints(_extension_name())


def get_importers() -> List[Type[Importer]]:
    """Get all available importers
    """
    return [e.plugin for e in get_import_mgr()]


def get_importer_by_name(name: str) -> Type[Importer]:
    """Get importer by name
    :param name: Name of the importer
    :type  name: str
    :returns: The importer
    :rtype:  Importer
    """
    imp = get_import_mgr()[name].plugin  # type: Type[Importer]
    return imp


def cache(fun: Callable[[Importer], Any]) -> Callable[[Importer], Any]:
    """
    This is a decorator to be used if a method of an Importer
    is to be cached, i.e., if the context of the importer is already
    set, then one does not need to run the function anymore, even if
    it is explicitly run.

    :param self: Method of an Importer
    """
    def wrapper(self: Importer) -> Any:
        if not self.ctx:
            fun(self)
    return wrapper



"""Module defining the main document type.
"""
import os
import datetime
import shutil
import logging
from typing import (
    List, Dict, Any, Optional, Union, NamedTuple, Callable, Tuple)
from typing_extensions import TypedDict

import papis.config
import papis

LOGGER = logging.getLogger("document")  # type: logging.Logger

KeyConversion = TypedDict(
    "KeyConversion", {"key": Optional[str],
                      "action": Optional[Callable[[Any], Any]]}
)
EmptyKeyConversion = {"key": None, "action": None}  # type: KeyConversion
KeyConversionPair = NamedTuple(
    "KeyConversionPair",
    [("foreign_key", str), ("list", List[KeyConversion])]
)


def keyconversion_to_data(conversion_list: List[KeyConversionPair],
                          data: Dict[str, Any],
                          keep_unknown_keys: bool = False) -> Dict[str, Any]:
    """Function to convert general dictionaries into a papis document.

    This can be used for instance when parsing a website and
    writing a KeyConversionPair to be input to this function.
    """

    new_data = dict()

    for key_pair in conversion_list:

        foreign_key = key_pair.foreign_key
        if foreign_key not in data:
            continue

        for conv_data in key_pair.list:
            papis_key = conv_data.get('key') or foreign_key  # type: str
            papis_value = data[foreign_key]

            try:
                action = conv_data.get('action') or (lambda x: x)
                new_data[papis_key] = action(papis_value)
            except Exception as ex:
                LOGGER.debug("Error while trying to parse %s (%s)",
                             papis_key, ex)

    if keep_unknown_keys:
        for key, value in data.items():
            if key in [c.foreign_key for c in conversion_list]:
                continue
            new_data[key] = value

    if 'author_list' in new_data:
        new_data['author'] = author_list_to_author(new_data)

    return new_data


def author_list_to_author(data: Dict[str, Any]) -> str:
    """Convert a list of authors into a single author string.
    """
    author = ''
    separator = papis.config.get('multiple-authors-separator')
    separator_fmt = papis.config.get('multiple-authors-format')
    if separator is None or separator_fmt is None:
        raise Exception(
            "You have to define 'multiple-author-separator'"
            " and 'multiple-author-format'")
    if 'author_list' in data:
        author = (
            separator.join([
                separator_fmt.format(au=author)
                for author in data['author_list']
            ])
        )
    return author


class DocHtmlEscaped(Dict[str, Any]):
    """
    Small helper class to escape html elements.

    >>> DocHtmlEscaped(from_data(dict(title='> >< int & "" "')))['title']
    '&gt; &gt;&lt; int &amp; &quot;&quot; &quot;'
    """

    def __init__(self, doc: Any) -> None:
        self.__doc = doc

    def __getitem__(self, key: str) -> str:
        return (
            str(self.__doc[key])
            .replace('&', '&amp;')
            .replace('<', '&lt;')
            .replace('>', '&gt;')
            .replace('"', '&quot;'))


class Document(Dict[str, Any]):

    """Class implementing the entry abstraction of a document in a library.
    It is basically a python dictionary with more methods.
    """

    subfolder = ""  # type: str
    _info_file_path = ""  # type: str

    def __init__(self, folder: Optional[str] = None,
                 data: Optional[Dict[str, Any]] = None):
        self._folder = None  # type: Optional[str]

        if folder is not None:
            self.set_folder(folder)
            self.load()

        if data is not None:
            self.update(data)

    def __missing__(self, key: str) -> str:
        """
        If key is not defined, return empty string
        """
        return ""

    @property
    def html_escape(self) -> DocHtmlEscaped:
        return DocHtmlEscaped(self)

    def get_main_folder(self) -> Optional[str]:
        """Get full path for the folder where the document and the information
        is stored.
        :returns: Folder path
        """
        return self._folder

    def set_folder(self, folder: str) -> None:
        """Set document's folder. The info_file path will be accordingly set.

        :param folder: Folder where the document will be stored, full path.
        :type  folder: str
        """
        self._folder = folder
        self._info_file_path = os.path.join(
            folder,
            papis.config.getstring('info-name'))
        self.subfolder = (self._folder
                              .replace(os.path.expanduser("~"), "")
                              .replace("/", " "))

    def get_main_folder_name(self) -> Optional[str]:
        """Get main folder name where the document and the information is
        stored.
        :returns: Folder name
        """
        folder = self.get_main_folder()
        return os.path.basename(folder) if folder else None

    def has(self, key: str) -> bool:
        """Check if the information file has some key defined.

        :param key: Key name to be checked
        :returns: True/False

        """
        return key in self

    def save(self) -> None:
        """Saves the current document's information into the info file.
        """
        # FIXME: fix circular import in papis.yaml
        import papis.yaml
        papis.yaml.data_to_yaml(self.get_info_file(),
                                {key: self[key]
                                 for key in self.keys() if self[key]})

    def get_info_file(self) -> str:
        """Get full path for the info file
        :returns: Full path for the info file
        :rtype: str
        """
        return self._info_file_path

    def get_files(self) -> List[str]:
        """Get the files linked to the document, if any.

        :returns: List of full file paths
        :rtype:  list
        """
        if not self.has('files'):
            return []
        files = (self["files"]
                 if isinstance(self["files"], list)
                 else [self["files"]])
        folder = self.get_main_folder()
        return [os.path.join(folder, fl) for fl in files] if folder else []

    def load(self) -> None:
        """Load information from info file
        """
        if not os.path.exists(self.get_info_file()):
            return
        try:
            data = papis.yaml.yaml_to_data(self.get_info_file(),
                                           raise_exception=True)
        except Exception as ex:
            LOGGER.error('Error reading yaml file in %s'
                         '\nPlease check it!\n\n%s',
                         self.get_info_file(), str(ex))
        else:
            for key in data:
                self[key] = data[key]


def from_folder(folder_path: str) -> Document:
    """Construct a document object from a folder

    :param folder_path: Full path to a valid papis folder
    :type  folder_path: str
    :returns: A papis document
    :rtype:  papis.document.Document
    """
    return Document(folder=folder_path)


def to_json(document: Document) -> str:
    """Export information into a json string
    :param document: Papis document
    :type  document: Document
    :returns: Json formatted info file
    :rtype:  str

    """
    import json
    return json.dumps(to_dict(document))


def to_dict(document: Document) -> Dict[str, Any]:
    """Gets a python dictionary with the information of the document
    :param document: Papis document
    :type  document: Document
    :returns: Python dictionary
    :rtype:  dict
    """
    result = dict()
    for key in document.keys():
        result[key] = document[key]
    return result


def dump(document: Document) -> str:
    """Return information string without any obvious format
    :param document: Papis document
    :type  document: Document
    :returns: String with document's information
    :rtype:  str

    >>> doc = from_data({'title': 'Hello World'})
    >>> dump(doc)
    'title:   Hello World\\n'
    """
    string = ""
    for i in document.keys():
        string += str(i)+":   "+str(document[i])+"\n"
    return string


def delete(document: Document) -> None:
    """This function deletes a document from disk.
    :param document: Papis document
    :type  document: papis.document.Document
    """
    folder = document.get_main_folder()
    if folder:
        shutil.rmtree(folder)


def describe(document: Union[Document, Dict[str, Any]]) -> str:
    """Return a string description of the current document
    using the document-description-format
    """
    return papis.format.format(
        papis.config.getstring('document-description-format'),
        document)


def move(document: Document, path: str) -> None:
    """This function moves a document to path, it supposes that
    the document exists in the location ``document.get_main_folder()``.
    Warning: This method will change the folder in the document object too.
    :param document: Papis document
    :type  document: papis.document.Document
    :param path: Full path where the document will be moved to
    :type  path: str

    >>> doc = from_data({'title': 'Hello World'})
    >>> doc.set_folder('path/to/folder')
    >>> import tempfile; newfolder = tempfile.mkdtemp()
    >>> move(doc, newfolder)
    Traceback (most recent call last):
    ...
    Exception: There is already...
    """
    path = os.path.expanduser(path)
    if os.path.exists(path):
        raise Exception(
            "There is already a document in {0}, please check it,\n"
            "a temporary papis document has been stored in {1}".format(
                path, document.get_main_folder()
            )
        )
    folder = document.get_main_folder()
    if folder:
        shutil.move(folder, path)
        # Let us chmod it because it might come from a temp folder
        # and temp folders are per default 0o600
        os.chmod(path, papis.config.getint('dir-umask') or 0o600)
        document.set_folder(path)


def from_data(data: Dict[str, Any]) -> Document:
    """Construct a document object from a data dictionary.

    :param data: Data to be copied to a new document
    :type  data: dict
    :returns: A papis document
    :rtype:  papis.document.Document
    """
    return Document(data=data)


def sort(docs: List[Document], key: str, reverse: bool) -> List[Document]:
    # The tuple returned by the _sort_for_key function represents:
    # (ranking, integer value, string value)
    # Rankings are:
    #   date:   0 (come first)
    #   integers:    2 (come after integers)
    #   strings:    3 (come after integers)
    #   None:       4 (come last)
    sort_rankings = {
        "date": 0,
        "int": 1,
        "string": 2,
        "None": 3
    }

    # Preserve the ordering of types even if --reverse is used.
    if reverse:
        for sort_type in sort_rankings:
            sort_rankings[sort_type] = -sort_rankings[sort_type]

    zero_date = datetime.datetime.fromtimestamp(0)

    def _sort_for_key(key: str, doc: Document
                      ) -> Tuple[int, datetime.datetime, int, str]:
        if key in doc.keys():
            if key == 'time-added':
                try:
                    date_value = \
                        datetime.datetime.strptime(str(doc[key]),
                                                   papis.strings.time_format)
                    return (sort_rankings["date"],
                            date_value, 0, str(doc[key]))
                except ValueError:
                    pass

            if str(doc[key]).isdigit():
                return (sort_rankings["int"],
                        zero_date,
                        int(doc[key]),
                        str(doc[key]))
            else:
                return (sort_rankings["string"],
                        zero_date,
                        0,
                        str(doc[key]))
        else:
            # The key does not appear in the document, ensure
            # it comes last.
            return (sort_rankings["None"], zero_date, 0, '')
    LOGGER.debug("sorting %d documents", len(docs))
    return sorted(docs, key=lambda d: _sort_for_key(key, d), reverse=reverse)


def new(folder_path: str, data: Dict[str, Any],
        files: List[str] = []) -> Document:
    """
    Creates a document at a given folder with data and
    some existing files.

    :param folder_path: A folder path, if non existing it will be created
    :type  folder_path: str
    :param data: Dictionary with key and values to be updated
    :type  data: dict
    :param files: Existing paths for files
    :type  files: list(str)
    :raises FileExistsError: If folder_path exists
    """
    os.makedirs(folder_path)
    doc = Document(folder=folder_path, data=data)
    doc['files'] = []
    for _file in files:
        shutil.copy(_file, os.path.join(folder_path))
        doc['files'].append(os.path.basename(_file))
    doc.save()
    return doc



import os

# Information
__license__ = 'GPLv3'
__version__ = '0.11.1'
__author__ = __maintainer__ = 'Alejandro Gallo'
__email__ = 'aamsgallo@gmail.com'


if os.environ.get('PAPIS_DEBUG'):
    import logging
    log_format = (
        '%(relativeCreated)d-' +
        '%(levelname)s' +
        ':' +
        '%(name)s' +
        ':' +
        '%(message)s'
    )
    logging.basicConfig(format=log_format, level=logging.DEBUG)



import re
from typing import Any, List, Dict, Union, Pattern
from typing_extensions import TypedDict

import bs4

import papis.bibtex
import papis.config
import papis.document
import papis.importer
import papis.utils


MetaEquivalence = TypedDict(
    "MetaEquivalence", {
        "tag": str,
        "key": str,
        "attrs": Dict[str, Union[str, Pattern[str]]],
    }
)

meta_equivalences = [
    # google
    {"tag": "meta", "key": "abstract", "attrs": {"name": "description"}},
    {"tag": "meta", "key": "doi", "attrs": {"name": "doi"}},
    {"tag": "meta", "key": "keywords", "attrs": {"name": "keywords"}},
    {"tag": "title", "key": "title", "attrs": {}},
    # facebook
    {"tag": "meta", "key": "type", "attrs": {"property": "og:type"}},
    {"tag": "meta",
        "key": "abstract", "attrs": {"property": "og:description"}},
    {"tag": "meta", "key": "title", "attrs": {"property": "og:title"}},
    {"tag": "meta", "key": "url", "attrs": {"property": "og:url"}},
    # citation style
    # https://scholar.google.com/intl/en/scholar/inclusion.html#indexing
    {"tag": "meta", "key": "doi", "attrs": {"name": "citation_doi"}},
    {"tag": "meta",
        "key": "firstpage", "attrs": {"name": "citation_firstpage"}},
    {"tag": "meta", "key": "lastpage", "attrs": {"name": "citation_lastpage"}},
    {"tag": "meta",
        "key": "url", "attrs": {"name": "citation_fulltext_html_url"}},
    {"tag": "meta", "key": "pdf_url", "attrs": {"name": "citation_pdf_url"}},
    {"tag": "meta", "key": "issn", "attrs": {"name": "citation_issn"}},
    {"tag": "meta", "key": "issue", "attrs": {"name": "citation_issue"}},
    {"tag": "meta", "key": "abstract", "attrs": {"name": "citation_abstract"}},
    {"tag": "meta",
        "key": "journal_abbrev", "attrs": {"name": "citation_journal_abbrev"}},
    {"tag": "meta",
        "key": "journal", "attrs": {"name": "citation_journal_title"}},
    {"tag": "meta", "key": "language", "attrs": {"name": "citation_language"}},
    {"tag": "meta",
        "key": "online_date", "attrs": {"name": "citation_online_date"}},
    {"tag": "meta",
        "key": "publication_date",
        "attrs": {"name": "citation_publication_date"}},
    {"tag": "meta",
        "key": "publisher", "attrs": {"name": "citation_publisher"}},
    {"tag": "meta", "key": "title", "attrs": {"name": "citation_title"}},
    {"tag": "meta", "key": "volume", "attrs": {"name": "citation_volume"}},
    # dc.{id} style
    {"tag": "meta",
        "key": "publisher",
        "attrs": {"name": re.compile("dc.publisher", re.I)}},
    {"tag": "meta",
        "key": "publisher",
        "attrs": {"name": re.compile(".*st.publisher.*", re.I)}},
    {"tag": "meta",
        "key": "date", "attrs": {"name": re.compile("dc.date", re.I)}},
    {"tag": "meta",
        "key": "language", "attrs": {"name": re.compile("dc.language", re.I)}},
    {"tag": "meta",
        "key": "issue",
        "attrs": {"name": re.compile("dc.citation.issue", re.I)}},
    {"tag": "meta",
            "key": "volume",
            "attrs": {"name": re.compile("dc.citation.volume", re.I)}},
    {"tag": "meta",
            "key": "keywords",
            "attrs": {"name": re.compile("dc.subject", re.I)}},
    {"tag": "meta",
            "key": "title", "attrs": {"name": re.compile("dc.title", re.I)}},
    {"tag": "meta",
            "key": "type", "attrs": {"name": re.compile("dc.type", re.I)}},
    {"tag": "meta",
            "key": "abstract",
            "attrs": {"name": re.compile("dc.description", re.I)}},
    {"tag": "meta",
            "key": "abstract",
            "attrs": {"name": re.compile("dc.description.abstract", re.I)}},
    {"tag": "meta",
            "key": "journal_abbrev",
            "attrs": {"name": re.compile("dc.relation.ispartof", re.I)}},
    {"tag": "meta",
            "key": "year", "attrs": {"name": re.compile("dc.issued", re.I)}},
    {"tag": "meta",
            "key": "doi",
            "attrs": {"name": re.compile("dc.identifier", re.I),
                      "scheme": "doi"}},
]  # type: List[MetaEquivalence]


def parse_meta_headers(soup: bs4.BeautifulSoup) -> Dict[str, Any]:
    global meta_equivalences
    # metas = soup.find_all(name="meta")
    data = dict()  # type: Dict[str, Any]
    for equiv in meta_equivalences:
        elements = soup.find_all(equiv['tag'], attrs=equiv["attrs"])
        if elements:
            value = elements[0].attrs.get("content")
            data[equiv["key"]] = str(value).replace('\r', '')

    author_list = parse_meta_authors(soup)
    if author_list:
        data['author_list'] = author_list
        data['author'] = papis.document.author_list_to_author(data)

    return data


def parse_meta_authors(soup: bs4.BeautifulSoup) -> List[Dict[str, Any]]:
    author_list = []  # type: List[Dict[str, Any]]
    authors = soup.find_all(name='meta', attrs={'name': 'citation_author'})
    if not authors:
        authors = soup.find_all(
            name='meta', attrs={'name': re.compile('dc.creator', re.I)})
    affs = soup.find_all(
        name='meta',
        attrs={'name': 'citation_author_institution'})
    if affs and authors:
        tuples = zip(authors, affs)
    elif authors:
        tuples = ((a, None) for a in authors)
    else:
        return []

    for t in tuples:
        fullname = t[0].get('content')
        affiliation = [dict(name=t[1].get('content'))] if t[1] else []
        fullnames = re.split(r'\s+', fullname)
        author_list.append(dict(
            given=fullnames[0],
            family=' '.join(fullnames[1:]),
            affiliation=affiliation,
        ))
    return author_list



"""
This command is to list contents of a library.

CLI Examples
^^^^^^^^^^^^

- List all document files associated will all entries:

    .. code:: bash

        papis list --all --file

    .. raw:: HTML

        <script type="text/javascript"
        src="https://asciinema.org/a/XwD0ZaUORoOonwDw4rXoQDkjZ.js"
        id="asciicast-XwD0ZaUORoOonwDw4rXoQDkjZ" async></script>

- List all document year and title with custom formatting:

    .. code:: bash

        papis list --all --format '{doc[year]} {doc[title]}'

    .. raw:: HTML

        <script type="text/javascript"
        src="https://asciinema.org/a/NZ8Ii1wWYPo477CIL4vZhUqOy.js"
        id="asciicast-NZ8Ii1wWYPo477CIL4vZhUqOy" async></script>

- List all documents according to the bibitem formatting (stored in a template
  file ``bibitem.template``):

    .. code:: bash

        papis list --all --template bibitem.template

    .. raw:: HTML

        <script type="text/javascript"
        src="https://asciinema.org/a/QZTBZ3tFfyk9WQuJ9WWB2UpSw.js"
        id="asciicast-QZTBZ3tFfyk9WQuJ9WWB2UpSw" async></script>

Cli
^^^
.. click:: papis.commands.list:cli
    :prog: papis list
"""

import logging
import papis
import os
import papis.utils
import papis.strings
import papis.config
import papis.database
import papis.document
import papis.downloaders
import papis.cli
import papis.pick
import papis.format
import click

from typing import List, Optional, Union, Sequence

logger = logging.getLogger('list')


def run(
        documents: List[papis.document.Document],
        libraries: bool = False,
        downloaders: bool = False,
        pick: bool = False,
        files: bool = False,
        folders: bool = False,
        info_files: bool = False,
        notes: bool = False,
        fmt: str = "",
        template: Optional[str] = None
        ) -> Sequence[Union[str, papis.document.Document]]:
    """Main method to the list command

    :returns: List different objects
    :rtype:  list
    """
    if downloaders:
        return [str(d) for d in papis.downloaders.get_available_downloaders()]

    if template is not None:
        if not os.path.exists(template):
            logger.error("Template file {} not found".format(template))
            return []
        with open(template) as fd:
            fmt = fd.read()

    if libraries:
        config = papis.config.get_configuration()
        return [
            section + ' ' + config[section]['dir']
            for section in config
            if 'dir' in config[section]]

    if files:
        return [
            doc_file for files in [
                document.get_files() for document in documents
            ] for doc_file in files
        ]
    elif notes:
        return [
            os.path.join(d.get_main_folder() or '', d["notes"])
            for d in documents
            if d.get_main_folder() is not None
            and d.has("notes") and isinstance(d["notes"], str)
            and os.path.exists(
                        os.path.join(d.get_main_folder() or '', d["notes"]))]
    elif info_files:
        return [d.get_info_file() for d in documents]
    elif fmt:
        return [
            papis.format.format(fmt, document)
            for document in documents
        ]
    elif folders:
        return [
            str(d.get_main_folder()) for d in documents
            if d.get_main_folder() is not None
        ]
    else:
        return documents


@click.command("list")
@click.help_option('--help', '-h')
@papis.cli.query_option()
@papis.cli.sort_option()
@click.option(
    "-i",
    "--info",
    help="Show the info file name associated with the document",
    default=False,
    is_flag=True)
@click.option(
    "-f", "--file", "_file",
    help="Show the file name associated with the document",
    default=False,
    is_flag=True)
@click.option(
    "-d", "--dir", "_dir",
    help="Show the folder name associated with the document",
    default=False,
    is_flag=True)
@click.option(
    "-n", "--notes",
    help="List notes files, if any",
    default=False,
    is_flag=True)
@click.option(
    "--format", "_format",
    help="List entries using a custom papis format, e.g."
    " '{doc[year] {doc[title]}",
    default='')
@click.option(
    "--template",
    help="Template file containing a papis format to list entries",
    default=None)
@click.option(
    "--downloaders",
    help="List available downloaders",
    default=False,
    is_flag=True)
@click.option(
    "--libraries",
    help="List defined libraries",
    default=False,
    is_flag=True)
@papis.cli.all_option()
def cli(
        query: str, info: bool, _file: bool, notes: bool, _dir: bool,
        _format: str,
        template: Optional[str], _all: bool, downloaders: bool,
        libraries: bool,
        sort_field: Optional[str], sort_reverse: bool) -> None:
    """List documents' properties"""

    logger = logging.getLogger('cli:list')
    documents = []  # type: List[papis.document.Document]

    if (not libraries and not downloaders and
            not _file and not info and not _dir):
        _dir = True

    if not libraries and not downloaders:
        db = papis.database.get()
        documents = db.query(query)
        if sort_field:
            documents = \
                papis.document.sort(documents, sort_field, sort_reverse)

        if not documents:
            logger.warning(papis.strings.no_documents_retrieved_message)
        if not _all:
            documents = list(papis.pick.pick_doc(documents))

    objects = run(
        documents,
        libraries=libraries,
        downloaders=downloaders,
        notes=notes,
        files=_file,
        folders=_dir,
        info_files=info,
        fmt=_format,
        template=template)

    for o in objects:
        click.echo(o)



import os
import glob
import papis.config
import papis.plugin
import re
from click.core import Command
from papis.cli import AliasedGroup
from typing import Dict, Optional, Union, NamedTuple


Script = NamedTuple("Script",
                    [("command_name", str),
                     ("path", Optional[str]),
                     ("plugin", Union[None, Command, AliasedGroup])])


def get_external_scripts() -> Dict[str, Script]:
    regex = re.compile('.*papis-([^ .]+)$')
    paths = []
    scripts = {}
    paths.append(papis.config.get_scripts_folder())
    paths += os.environ["PATH"].split(":")
    for path in paths:
        for script in glob.glob(os.path.join(path, "papis-*")):
            m = regex.match(script)
            if m is not None:
                name = m.group(1)
                scripts[name] = Script(command_name=name,
                                       path=script,
                                       plugin=None)
    return scripts


def _extension_name() -> str:
    return "papis.command"


def get_scripts() -> Dict[str, Script]:
    commands_mgr = papis.plugin.get_extension_manager(_extension_name())
    scripts_dict = dict()
    for command_name in commands_mgr.names():
        scripts_dict[command_name] = Script(
            command_name=command_name,
            path=None,
            plugin=commands_mgr[command_name].plugin
        )
    return scripts_dict



"""
General architecture
--------------------

Papis uses the package
`stevedore <https://github.com/openstack/stevedore/>`_
for general plugin management.

The only papis module invoking ``stevedore`` should be
``papis/plugin.py``.

The different plugins in papis like ``papis.command``,
``papis.exporter`` etc. define a so-called ``ExtensionManager``
which loads various objects that have been declared in a python
package somewhere.

For example, the yaml exporter is defined as

.. code:: python

    def exporter(documents: List[papis.document.Document]) -> str:
        string = yaml.dump_all(
            [papis.document.to_dict(document) for document in documents],
            allow_unicode=True)
        return str(string)

and declared in ``setup.py`` as

.. code:: python

    setup(
        ...
        entry_points={
            'papis.exporter': [
                'bibtex=papis.bibtex:exporter',
                'json=papis.json:exporter',
                'yaml=papis.yaml:exporter',
            ],
        ...
    )

and the exporter can be used as in the code below

.. code:: python

    import papis.plugin
    extension_manager = papis.plugin.get_extension_manager("papis.exporter")
    # yaml_exporter is the function defined above
    yaml_exporter = extension_manager["yaml"].plugin

    yaml_string = yaml_exporter(mydocs)

Now a developer is able to write another exporter in some package
and install the package in the system.
The ``extension_manager`` will be able to access the provided functions
in the package if they have been declared in the entry points of
the ``setup.py`` script of the named package.
"""
import logging
from stevedore import ExtensionManager
from typing import List, Dict, Any  # noqa: ignore

logger = logging.getLogger("papis:plugin")


MANAGERS = dict()  # type: Dict[str, ExtensionManager]


def stevedore_error_handler(manager: ExtensionManager,
                            entrypoint: str, exception: str) -> None:
    logger.error("Error while loading entrypoint [%s]" % entrypoint)
    logger.error(exception)


def _load_extensions(namespace: str) -> None:
    global MANAGERS
    logger.debug("creating manager for {0}".format(namespace))
    MANAGERS[namespace] = ExtensionManager(
        namespace=namespace,
        invoke_on_load=False,
        verify_requirements=True,
        propagate_map_exceptions=True,
        on_load_failure_callback=stevedore_error_handler
    )


def get_extension_manager(namespace: str) -> ExtensionManager:
    global MANAGERS
    if not MANAGERS.get(namespace):
        _load_extensions(namespace)
    extension_mgr = MANAGERS[namespace]
    assert extension_mgr is not None
    return extension_mgr


def get_available_entrypoints(namespace: str) -> List[str]:
    return list(
        map(str, get_extension_manager(namespace).entry_points_names()))


def get_available_plugins(namespace: str) -> List[Any]:
    return [e.plugin for e in get_extension_manager(namespace)]



"""This command edits the information of the documents.
The editor used is defined by the ``editor`` configuration setting.


Cli
^^^
.. click:: papis.commands.edit:cli
    :prog: papis edit
"""
import papis
import os
import papis.api
import papis.pick
import papis.document
import papis.utils
import papis.config
import papis.database
import papis.cli
import click
import logging
import papis.strings
import papis.git

from typing import Optional


def run(document: papis.document.Document,
        wait: bool = True,
        git: bool = False) -> None:
    database = papis.database.get()
    info_file_path = document.get_info_file()
    if not info_file_path:
        raise Exception(papis.strings.no_folder_attached_to_document)
    papis.utils.general_open(info_file_path, "editor", wait=wait)
    document.load()
    database.update(document)
    if git:
        papis.git.add_and_commit_resource(
            str(document.get_main_folder()),
            info_file_path,
            "Update information for '{0}'".format(
                papis.document.describe(document)))


@click.command("edit")
@click.help_option('-h', '--help')
@papis.cli.query_option()
@papis.cli.doc_folder_option()
@papis.cli.git_option(help="Add changes made to the info file")
@papis.cli.sort_option()
@click.option(
    "-n",
    "--notes",
    help="Edit notes associated to the document",
    default=False,
    is_flag=True)
@papis.cli.all_option()
@click.option(
    "-e",
    "--editor",
    help="Editor to be used",
    default=None)
def cli(query: str,
        doc_folder: str,
        git: bool,
        notes: bool,
        _all: bool,
        editor: Optional[str],
        sort_field: Optional[str],
        sort_reverse: bool) -> None:
    """Edit document information from a given library"""

    logger = logging.getLogger('cli:edit')

    if doc_folder:
        documents = [papis.document.from_folder(doc_folder)]
    else:
        documents = papis.database.get().query(query)

    if sort_field:
        documents = papis.document.sort(documents, sort_field, sort_reverse)

    if editor is not None:
        papis.config.set('editor', editor)

    if not _all:
        documents = list(papis.pick.pick_doc(documents))

    if len(documents) == 0:
        logger.warning(papis.strings.no_documents_retrieved_message)
        return

    for document in documents:
        if notes:
            logger.debug("Editing notes")
            if not document.has("notes"):
                logger.warning(
                    "The document selected has no notes attached, \n"
                    "creating a notes files"
                )
                document["notes"] = papis.config.getstring("notes-name")
                document.save()
            notes_path = os.path.join(
                str(document.get_main_folder()),
                document["notes"]
            )

            if not os.path.exists(notes_path):
                logger.info("Creating {0}".format(notes_path))
                open(notes_path, "w+").close()

            papis.api.edit_file(notes_path)
            if git:
                papis.git.add_and_commit_resource(
                    str(document.get_main_folder()),
                    str(document.get_info_file()),
                    "Update notes for '{0}'".format(
                        papis.document.describe(document)))

        else:
            run(document, git=git)



import sys
import os
from os.path import expanduser
import configparser
import papis.exceptions
import papis.library
import logging
from typing import Dict, Any, List, Optional, Callable  # noqa: ignore

PapisConfigType = Dict[str, Dict[str, Any]]

logger = logging.getLogger("config")
logger.debug("importing")

_CURRENT_LIBRARY = None  #: Current library in use
_CONFIGURATION = None  # type: Optional[Configuration]
_DEFAULT_SETTINGS = None  # type: Optional[PapisConfigType]
_OVERRIDE_VARS = {
    "folder": None,
    "file": None,
    "scripts": None
}  # type: Dict[str, Optional[str]]


def get_default_opener() -> str:
    """Get the default file opener for the current system
    """
    if sys.platform.startswith('darwin'):
        return "open"
    elif os.name == 'nt':
        return "start"
    else:
        return "xdg-open"


general_settings = {
    "local-config-file": ".papis.config",
    "database-backend": "papis",
    "default-query-string": ".",
    "sort-field": None,

    "opentool": get_default_opener(),
    "dir-umask": 0o755,
    "browser": os.environ.get('BROWSER') or get_default_opener(),
    "picktool": "papis",
    "mvtool": "mv",
    "editor": os.environ.get('EDITOR')
                        or os.environ.get('VISUAL')
                        or get_default_opener(),
    "notes-name": "notes.tex",
    "use-cache": True,
    "cache-dir": None,
    "use-git": False,

    "add-confirm": False,
    "add-folder-name": "",
    "add-file-name": None,
    "add-interactive": False,
    "add-edit": False,
    "add-open": False,

    "browse-key": 'url',
    "browse-query-format": "{doc[title]} {doc[author]}",
    "search-engine": "https://duckduckgo.com",
    "user-agent": 'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_9_3)',
    "scripts-short-help-regex": ".*papis-short-help: *(.*)",
    "info-name": "info.yaml",
    "doc-url-key-name": "doc_url",

    "open-mark": False,
    "mark-key-name": "marks",
    "mark-format-name": "mark",
    "mark-header-format": "{mark[name]} - {mark[value]}",
    "mark-match-format": "{mark[name]} - {mark[value]}",
    "mark-opener-format": get_default_opener(),

    "file-browser": get_default_opener(),
    "bibtex-journal-key": 'journal',
    "extra-bibtex-keys": "[]",
    "extra-bibtex-types": "[]",
    "default-library": "papers",
    "format-doc-name": "doc",
    "match-format":
        "{doc[tags]}{doc.subfolder}{doc[title]}{doc[author]}{doc[year]}",
    "header-format-file": None,
    "header-format": (
        "<ansired>{doc.html_escape[title]}</ansired>\n"
        " <ansigreen>{doc.html_escape[author]}</ansigreen>\n"
        "  <blue>({doc.html_escape[year]})</blue> "
        "[<ansiyellow>{doc.html_escape[tags]}</ansiyellow>]"
    ),

    "info-allow-unicode": True,
    "ref-format": "{doc[title]:.15} {doc[author]:.6} {doc[year]}",
    "multiple-authors-separator": " and ",
    "multiple-authors-format": "{au[family]}, {au[given]}",

    "whoosh-schema-fields": "['doi']",
    "whoosh-schema-prototype":
    '{\n'
    '"author": TEXT(stored=True),\n'
    '"title": TEXT(stored=True),\n'
    '"year": TEXT(stored=True),\n'
    '"tags": TEXT(stored=True),\n'
    '}',

    'unique-document-keys': "['doi','ref','isbn','isbn10','url','doc_url']",

    "downloader-proxy": None,
    "bibtex-unicode": False,

    "time-stamp": True,

    "document-description-format": '{doc[title]} - {doc[author]}',
    "formater": "python",
}


def get_general_settings_name() -> str:
    """Get the section name of the general settings
    :returns: Section's name
    :rtype:  str
    >>> get_general_settings_name()
    'settings'
    """
    return "settings"


class Configuration(configparser.ConfigParser):

    def __init__(self) -> None:
        configparser.ConfigParser.__init__(self)
        self.dir_location = get_config_folder()
        self.scripts_location = get_scripts_folder()
        self.file_location = get_config_file()
        self.logger = logging.getLogger("Configuration")
        self.default_info = {
          "papers": {
            'dir': '~/Documents/papers'
          },
          get_general_settings_name(): {
            'default-library': 'papers'
          }
        }  # type: PapisConfigType
        self.initialize()

    def handle_includes(self) -> None:
        if "include" in self.keys():
            for name in self["include"]:
                self.logger.debug("including %s" % name)
                fullpath = os.path.expanduser(self.get("include", name))
                if os.path.exists(fullpath):
                    self.read(fullpath)
                else:
                    self.logger.warning(
                        "{0} not included because it does not exist"
                        .format(fullpath))

    def initialize(self) -> None:
        if not os.path.exists(self.dir_location):
            self.logger.warning(
                'Creating configuration folder in {0}'
                .format(self.dir_location))
            os.makedirs(self.dir_location)
        if not os.path.exists(self.scripts_location):
            os.makedirs(self.scripts_location)
        if os.path.exists(self.file_location):
            self.logger.debug(
                'Reading configuration from {0}'.format(self.file_location))
            self.read(self.file_location)
            self.handle_includes()
        else:
            for section in self.default_info:
                self[section] = {}
                for field in self.default_info[section]:
                    self[section][field] = self.default_info[section][field]
            with open(self.file_location, "w") as configfile:
                self.logger.info('Creating config file at {0}'
                                 .format(self.file_location))
                self.write(configfile)
        configpy = get_configpy_file()
        if os.path.exists(configpy):
            self.logger.debug('Executing {0}'.format(configpy))
            with open(configpy) as fd:
                exec(fd.read())


def get_default_settings() -> PapisConfigType:
    """Get the default settings for all non-user variables
    in papis.

    """
    global _DEFAULT_SETTINGS
    # We use an OrderedDict so that the first entry will always be the general
    # settings, also good for automatic documentation
    if _DEFAULT_SETTINGS is None:
        _DEFAULT_SETTINGS = dict()
        _DEFAULT_SETTINGS.update({
            get_general_settings_name(): general_settings,
        })
        import papis.tui
        _DEFAULT_SETTINGS.update(papis.tui.get_default_settings())
    return _DEFAULT_SETTINGS


def register_default_settings(settings_dictionary: PapisConfigType) -> None:
    """Register configuration settings into the global configuration registry.

    Notice that you can define sections or global options. For instance,
    let us suppose that a script called ``hubation`` defines some
    configuration options. In the script there could be the following defined

    ::

        import papis.config
        options = {'hubation': { 'command': 'open'}}
        papis.config.register_default_settings(options)

    and later on the script can use these options as:

    ::

        papis.config.get('command', section='hubation')

    :param settings_dictionary: A dictionary with settings
    :type  settings_dictionary: dict
    """
    default_settings = get_default_settings()
    # we do a for loop because apparently the OrderedDict removes all
    # key-val fields after updating, so we have to do it by hand
    for section in settings_dictionary.keys():
        if section in default_settings.keys():
            default_settings[section].update(settings_dictionary[section])
        else:
            default_settings[section] = settings_dictionary[section]


def get_config_home() -> str:
    """Returns the base directory relative to which user specific configuration
    files should be stored.

    :returns: Configuration base directory
    :rtype:  str
    """
    xdg_home = os.environ.get('XDG_CONFIG_HOME')
    if xdg_home:
        return expanduser(xdg_home)
    else:
        return os.path.join(expanduser('~'), '.config')


def get_config_dirs() -> List[str]:
    """Get papis configuration directories where the configuration
    files might be stored
    """
    dirs = []  # type: List[str]
    # get_config_home should also be included on top of XDG_CONFIG_DIRS
    if os.environ.get('XDG_CONFIG_DIRS') is not None:
        dirs += [
            os.path.join(d, 'papis') for d in
            os.environ.get('XDG_CONFIG_DIRS', '').split(':')]
    # Take XDG_CONFIG_HOME and $HOME/.papis for backwards
    # compatibility
    dirs += [
        os.path.join(get_config_home(), 'papis'),
        os.path.join(expanduser('~'), '.papis')]
    return dirs


def get_config_folder() -> str:
    """Get folder where the configuration files are stored,
    e.g. ``/home/user/.papis``. It is XDG compatible, which means that if the
    environment variable ``XDG_CONFIG_HOME`` is defined it will use the
    configuration folder ``XDG_CONFIG_HOME/papis`` instead.

    """
    config_dirs = get_config_dirs()
    for config_dir in config_dirs:
        if os.path.exists(config_dir):
            return config_dir
    # If no folder is found, then get the config home
    return os.path.join(get_config_home(), 'papis')


def get_config_file() -> str:
    """Get the path of the main configuration file,
    e.g. /home/user/.papis/config
    """
    global _OVERRIDE_VARS
    if _OVERRIDE_VARS["file"] is not None:
        config_file = _OVERRIDE_VARS["file"]
    else:
        config_file = os.path.join(get_config_folder(), "config")
    logger.debug("Getting config file %s" % config_file)
    return config_file


def get_configpy_file() -> str:
    """Get the path of the main python configuration file,
    e.g. /home/user/config/.papis/config.py
    """
    return os.path.join(get_config_folder(), "config.py")


def set_config_file(filepath: str) -> None:
    """Override the main configuration file path
    """
    global _OVERRIDE_VARS
    logger.debug("Setting config file to %s" % filepath)
    _OVERRIDE_VARS["file"] = filepath


def get_scripts_folder() -> str:
    """Get folder where the scripts are stored,
    e.g. /home/user/.papis/scripts
    """
    return os.path.join(get_config_folder(), "scripts")


def set(key: str, val: Any, section: Optional[str] = None) -> None:
    """Set a key to val in some section and make these changes available
    everywhere.
    """
    config = get_configuration()
    if not config.has_section(section or "settings"):
        config.add_section(section or "settings")
    config[section or get_general_settings_name()][key] = str(val)


def general_get(key: str, section: Optional[str] = None,
                data_type: Optional[Any] = None) -> Optional[Any]:
    """General getter method that will be specialized for different modules.

    :param data_type: The data type that should be expected for the value of
        the variable.
    :type  data_type: DataType, e.g. int, src ...
    :param default: Default value for the configuration variable if it is not
        set.
    :type  default: It should be the same that ``data_type``
    :param extras: List of tuples containing section and prefixes
    """
    # Init main variables
    method = None  # type: Optional[Callable[[Any, Any], Any]]
    value = None  # type: Optional[Any]
    config = get_configuration()
    libname = get_lib_name()
    global_section = get_general_settings_name()
    specialized_key = section + "-" + key if section is not None else key
    extras = [(section, key)] if section is not None else []
    sections = [(global_section, specialized_key)] +\
        extras + [(libname, specialized_key)]
    default_settings = get_default_settings()

    # Check data type for setting getter method
    if data_type == int:
        method = config.getint
    elif data_type == float:
        method = config.getfloat
    elif data_type == bool:
        method = config.getboolean
    else:
        method = config.get

    # Try to get key's value from configuration
    for extra in sections:
        sec = extra[0]
        whole_key = extra[1]
        if sec not in config.keys():
            continue
        if whole_key in config[sec].keys():
            value = method(sec, whole_key)

    if value is None:
        try:
            default = default_settings[
                section or global_section
            ][
                specialized_key if section is None else key
            ]
        except KeyError:
            raise papis.exceptions.DefaultSettingValueMissing(key)
        else:
            return default
    return value


def get(key: str, section: Optional[str] = None) -> Optional[Any]:
    """String getter
    """
    return general_get(key, section=section)


def getint(key: str, section: Optional[str] = None) -> Optional[int]:
    """Integer getter
    >>> set('something', 42)
    >>> getint('something')
    42
    """
    return general_get(key, section=section, data_type=int)


def getfloat(key: str, section: Optional[str] = None) -> Optional[float]:
    """Float getter
    >>> set('something', 0.42)
    >>> getfloat('something')
    0.42
    """
    return general_get(key, section=section, data_type=float)


def getboolean(key: str, section: Optional[str] = None) -> Optional[bool]:
    """Bool getter
    >>> set('add-open', True)
    >>> getboolean('add-open')
    True
    """
    return general_get(key, section=section, data_type=bool)


def getstring(key: str, section: Optional[str] = None) -> str:
    """String getter
    >>> set('add-open', "hello world")
    >>> getstring('add-open')
    'hello world'
    """
    result = general_get(key, section=section, data_type=str)
    if not isinstance(result, str):
        raise ValueError("Key {0} should be a string".format(key))
    return str(result)


def getlist(key: str, section: Optional[str] = None) -> List[str]:
    """List getter

    :returns: A python list
    :rtype:  list
    :raises SyntaxError: Whenever the parsed syntax is either not a valid
        python object or a valid python list.
    """
    rawvalue = general_get(key, section=section)  # type: Any
    if isinstance(rawvalue, list):
        return list(map(str, rawvalue))
    try:
        rawvalue = eval(rawvalue)
    except Exception as e:
        raise SyntaxError(
            "The key '{0}' must be a valid python object\n\t{1}"
            .format(key, e))
    else:
        if not isinstance(rawvalue, list):
            raise SyntaxError(
                "The key '{0}' must be a valid python list".format(key))
        return list(map(str, rawvalue))


def get_configuration() -> Configuration:
    """Get the configuration object, if no papis configuration has ever been
    initialized, it initializes one. Only one configuration per process should
    ever be configured.

    :returns: Configuration object
    :rtype:  papis.config.Configuration
    """
    global _CONFIGURATION
    if _CONFIGURATION is None:
        logger.debug("Creating configuration")
        _CONFIGURATION = Configuration()
        # Handle local configuration file, and merge it if it exists
        local_config_file = papis.config.get("local-config-file")
        merge_configuration_from_path(local_config_file, _CONFIGURATION)
    return _CONFIGURATION


def merge_configuration_from_path(path: Optional[str],
                                  configuration: Configuration) -> None:
    """
    Merge information of a configuration file found in `path`
    to the information of the configuration object stored in `configuration`.

    :param path: Path to the configuration file
    :type  path: str
    :param configuration: Configuration object
    :type  configuration: papis.config.Configuration
    """
    if path is None or not os.path.exists(path):
        return
    logger.debug("Merging configuration from " + path)
    configuration.read(path)
    configuration.handle_includes()


def set_lib(library: papis.library.Library) -> None:
    """Set library

    :param library: Library object
    :type  library: papis.library.Library

    """
    global _CURRENT_LIBRARY
    config = get_configuration()
    if library.name not in config.keys():
        config[library.name] = dict(dirs=str(library.paths))
    _CURRENT_LIBRARY = library


def set_lib_from_name(libname: str) -> None:
    """Set library, notice that in principle library can be a full path.

    :param libname: Name of the library or some path to a folder
    :type  libname: str
    """
    set_lib(get_lib_from_name(libname))


def get_lib_from_name(libname: str) -> papis.library.Library:
    config = get_configuration()
    if libname not in config.keys():
        if os.path.isdir(libname):
            # Check if the path exists, then use this path as a new library
            logger.warning("Since the path '{0}' exists, "
                           "I'm interpreting it as a library"
                           .format(libname))
            library_obj = papis.library.from_paths([libname])
            name = library_obj.path_format()
            # the configuration object can only store strings
            config[name] = dict(dirs=str(library_obj.paths))
        else:
            raise Exception("Library '{0}' does not seem to exist"
                            "\n\n"
                            "To add a library simply write the following"
                            "in your configuration file located at '{cpath}'"
                            "\n\n"
                            "\t[{0}]\n"
                            "\tdir = path/to/your/{0}/folder"
                            .format(libname, cpath=get_config_file()))
    else:
        try:
            paths = [expanduser(config[libname]['dir'])]
        except KeyError:
            try:
                paths = eval(expanduser(config[libname].get('dirs')))
            except Exception as e:
                raise Exception("To define a library you have to set either"
                                " dir or dirs in the configuration file.\n"
                                "\tdir must be a path to a single folder.\n"
                                "\tdirs must be a python list of "
                                "paths to folders.\n\n"
                                "Error: ({0})"
                                .format(e))
        library_obj = papis.library.Library(libname, paths)
    return library_obj


def get_lib_dirs() -> List[str]:
    """Get the directories of the current library

    :returns: A list of paths
    :rtype:  list
    """
    return get_lib().paths


def get_lib_name() -> str:
    return get_lib().name


def get_lib() -> papis.library.Library:
    """Get current library, if there is no library set before,
    the default library will be retrieved.
    If the `PAPIS_LIB` environment variable is defined, this is the
    library name (or path) that will be taken as a default.

    :returns: Current library
    :rtype:  papis.library.Library
    """
    global _CURRENT_LIBRARY
    if os.environ.get('PAPIS_LIB'):
        set_lib_from_name(os.environ['PAPIS_LIB'])
    if _CURRENT_LIBRARY is None:
        # Do not put papis.config.get because get is a special function
        # that also needs the library to see if some key was overridden!
        default_settings = get_default_settings()[get_general_settings_name()]
        lib = default_settings['default-library']
        set_lib_from_name(lib)
    assert(isinstance(_CURRENT_LIBRARY, papis.library.Library))
    return _CURRENT_LIBRARY


def reset_configuration() -> Configuration:
    """Destroys existing configuration and returns a new one.

    :returns: Configuration object
    :rtype:  papis.config.Configuration
    """
    global _CONFIGURATION
    _CONFIGURATION = None
    logger.debug("Resetting configuration")
    return get_configuration()



"""This module serves as an lightweight interface for git related functions.
"""
import subprocess
import os
import shlex
import logging

LOGGER = logging.getLogger("papis.git")


def _issue_git_command(path: str, cmd: str) -> None:
    """Issues a general git command ``cmd`` at ``path``.

    :param path: Folder where a git repo exists.
    :type  path: str
    :param message: Git command
    :type  message: str
    :returns: None

    """
    assert isinstance(cmd, str)
    path = os.path.expanduser(path)
    assert os.path.exists(path)
    split_cmd = shlex.split(cmd)
    os.chdir(path)
    LOGGER.debug(split_cmd)
    subprocess.call(split_cmd)


def commit(path: str, message: str) -> None:
    """Commits changes in the path with a message.

    :param path: Folder where a git repo exists.
    :type  path: str
    :param message: Commit message
    :type  message: str
    :returns: None

    """
    LOGGER.info('Commiting %s with message %s', path, message)
    cmd = 'git commit -m "{0}"'.format(message)
    _issue_git_command(path, cmd)


def add(path: str, resource: str) -> None:
    """Adds changes in the path with a message.

    :param path: Folder where a git repo exists.
    :type  path: str
    :param resource: Commit resource
    :type  resource: str
    :returns: None

    """
    LOGGER.info('Adding %s', path)
    cmd = 'git add "{0}"'.format(resource)
    _issue_git_command(path, cmd)


def remove(path: str, resource: str, recursive: bool = False) -> None:
    """Adds changes in the path with a message.

    :param path: Folder where a git repo exists.
    :type  path: str
    :param resource: Commit resource
    :type  resource: str
    :returns: None

    """
    LOGGER.info('Removing %s', path)
    # force removal always
    cmd = 'git rm -f {r} "{0}"'.format(resource, r="-r" if recursive else "")
    _issue_git_command(path, cmd)


def add_and_commit_resource(path: str, resource: str, message: str) -> None:
    """Adds and commits a resource.
    """
    add(path, resource)
    commit(path, message)



import doi
import papis.downloaders.base

from typing import Dict, Any, Optional


class Downloader(papis.downloaders.Downloader):

    def __init__(self, uri: str, name: str = "fallback"):
        papis.downloaders.Downloader.__init__(
            self, uri=uri, name=name)
        self.priority = -1

    @classmethod
    def match(cls, url: str) -> Optional[papis.downloaders.Downloader]:
        return Downloader(url)

    def get_data(self) -> Dict[str, Any]:
        data = dict()
        soup = self._get_soup()
        data.update(papis.downloaders.base.parse_meta_headers(soup))
        return data

    def get_doi(self) -> Optional[str]:
        if self.ctx.data and 'doi' in self.ctx.data:
            _doi = self.ctx.data['doi']
            return str(_doi) if _doi else None
        soup = self._get_soup()
        self.logger.info('trying to parse doi from url body...')
        if soup:
            return doi.find_doi_in_text(str(soup))
        else:
            return None

    def get_document_url(self) -> Optional[str]:
        if 'pdf_url' in self.ctx.data:
            url = self.ctx.data.get('pdf_url')
            self.logger.debug("got a pdf url = %s" % url)
            return url
        else:
            return None



"""
This command is useful if your library is itself a git repository.
You can use this command to issue git commands in your library
repository without having to change your current directory.

CLI Examples
^^^^^^^^^^^^

    - Check the status of the library repository:

    .. code::

        papis git status

    - Commit all changes:

    .. code::

        papis git commit -a


"""
import copy
import click
import papis.commands.run

cli = copy.deepcopy(papis.commands.run.cli)
cli.help = "Run git command in a library or document folder"
cli = click.option("--prefix", hidden=True, default="git", type=str)(cli)



"""This command will try its best to find a source in the internet for the
document at hand.

Of course if the document has an url key in its info file, it will use this url
to open it in a browser.  Also if it has a doc_url key, or a doi, it will try
to compose urls out of these to open it.

If none of the above work, then it will try to use a search engine with the
document's information (using the ``browse-query-format``).  You can select
wich search engine you want to use using the ``search-engine`` setting.

It uses the configuration option ``browse-key`` to form an url
according to which key is given in the document. You can bypass this option
using the `-k` flag issuing the command.

::

    papis browse -k doi einstein

This will form an url through the DOI of the document.

::

    papis browse -k isbn

This will form an url through the ISBN of the document
using isbnsearch.org.

::

    papis browse -k ads

This will form an url using the gread ADS service and there you can check
for similar papers, citations, references and much more.
Please note that for this to work the document should have a DOI
attached to it.

::

    papis browse -k whatever

This will consider the key ``whatever`` of the document
to be a valid url, I guess at this point you'll know what you're doing.

::

    papis browse -k search-engine

This is the default, it will do a search-engine search with the data of your
paper and hopefully you'll find it.

Cli
^^^
.. click:: papis.commands.browse:cli
    :prog: papis browse
"""
import papis
import papis.utils
import papis.config
import papis.cli
import papis.pick
import click
import papis.database
import papis.strings
import papis.document
from urllib.parse import urlencode
import logging

from typing import Optional

logger = logging.getLogger('browse')


def run(document: papis.document.Document) -> Optional[str]:
    """Browse document's url whenever possible and returns the url

    :document: Document object

    """
    global logger
    url = None
    key = papis.config.getstring("browse-key")

    try:
        if "doi" == key:
            url = 'https://doi.org/{}'.format(document['doi'])
        elif "ads" == key:
            url = ('https://ui.adsabs.harvard.edu/abs/%22{}%22'
                   .format(document['doi']))
        elif "isbn" == key:
            url = 'https://isbnsearch.org/isbn/{}'.format(document['isbn'])
        else:
            url = document[key]

        if not url:
            raise KeyError()

    except KeyError:
        if not url or key == 'search-engine':
            params = {
                'q': papis.format.format(
                    papis.config.getstring('browse-query-format'),
                    document
                )
            }
            url = (papis.config.getstring('search-engine')
                   + '/?'
                   + urlencode(params)
                   )

    logger.info("Opening url %s:" % url)
    papis.utils.general_open(url, "browser", wait=False)
    return url


@click.command("browse")
@click.help_option('--help', '-h')
@papis.cli.query_option()
@papis.cli.sort_option()
@click.option('-k', '--key', default='',
              help='Use the value of the document\'s key to open in'
                   ' the browser, e.g. doi, url, doc_url ...')
@papis.cli.all_option()
@papis.cli.doc_folder_option()
def cli(query: str,
        key: str,
        _all: bool,
        doc_folder: str,
        sort_field: Optional[str],
        sort_reverse: bool) -> None:
    """Open document's url in a browser"""

    if doc_folder:
        documents = [papis.document.from_folder(doc_folder)]
    else:
        documents = papis.database.get().query(query)

    logger = logging.getLogger('cli:browse')

    if len(documents) == 0:
        logger.warning(papis.strings.no_documents_retrieved_message)
        return

    if not _all:
        documents = list(papis.pick.pick_doc(documents))
        if not documents:
            return

    if sort_field:
        documents = papis.document.sort(documents, sort_field, sort_reverse)

    if len(key):
        papis.config.set('browse-key', key)

    logger.info("Using key = %s", papis.config.get("browse-key"))

    for document in documents:
        run(document)



"""

Examples
^^^^^^^^

- To override some configuration options, you can use the flag ``--set``, for
  instance, if you want to override the editor used and the opentool to open
  documents, you can just type

    .. code:: shell

        papis --set editor gedit --set opentool firefox edit
        papis --set editor gedit --set opentool firefox open

- If you want to list the libraries and pick one before sending a database
  query to papis, use ``--pick-lib`` as such

    .. code:: shell

        papis --pick-lib open 'einstein relativity'

Cli
^^^
.. click:: papis.commands.default:run
    :prog: papis
    :commands: []

"""
import os
import difflib
import sys
import logging
from typing import Optional, Tuple, List, Callable

import papis
import papis.api
import papis.config
import papis.commands
import papis.database
import papis.cli

import atexit
import cProfile
import pstats
import colorama
import click
import click.core


class MultiCommand(click.core.MultiCommand):

    scripts = papis.commands.get_scripts()
    scripts.update(papis.commands.get_external_scripts())
    logger = logging.getLogger('multicommand')

    def list_commands(self, ctx: click.core.Context) -> List[str]:
        """List all matched commands in the command folder and in path

        >>> mc = MultiCommand()
        >>> rv = mc.list_commands(None)
        >>> len(rv) > 0
        True
        """
        _rv = list(self.scripts.keys())
        _rv.sort()
        return _rv

    def get_command(
            self,
            ctx: click.core.Context,
            name: str) -> Optional[click.core.Command]:
        """Get the command to be run

        >>> mc = MultiCommand()
        >>> cmd = mc.get_command(None, 'add')
        >>> cmd.name, cmd.help
        ('add', 'Add...')
        >>> mc.get_command(None, 'this command does not exist')
        """
        try:
            script = self.scripts[name]
        except KeyError:
            matches = list(map(
                str, difflib.get_close_matches(name, self.scripts, n=2)))
            self.logger.error(
                '{c.Fore.RED}{c.Style.BRIGHT}{c.Back.BLACK}'
                'did you mean {0}?'
                '{c.Style.RESET_ALL}'
                .format(
                    ' or '.join(matches),
                    c=colorama
                ))
            # return the match if there was only one match
            if len(matches) == 1:
                self.logger.warning("I suppose you meant: '%s'", *matches)
                script = self.scripts[matches[0]]
            else:
                return None

        if script.plugin is not None:
            return script.plugin
        # If it gets here, it means that it is an external script
        from papis.commands.external import external_cli as cli
        from papis.commands.external import get_command_help
        cli.context_settings['obj'] = script
        if script.path is not None:
            cli.help = get_command_help(script.path)
        cli.name = script.command_name
        cli.short_help = cli.help
        return cli


def generate_profile_writing_function(profiler: cProfile.Profile,
                                      filename: str) -> Callable[[], None]:
    def _on_finish() -> None:
        profiler.disable()
        profiler.create_stats()
        with open(filename, 'w') as output:
            stats = pstats.Stats(profiler, stream=output)
            stats.sort_stats('time')
            stats.print_stats()

    return _on_finish


@click.group(
    cls=MultiCommand,
    invoke_without_command=True)
@click.help_option('--help', '-h')
@click.version_option(version=papis.__version__)
@click.option(
    "-v",
    "--verbose",
    help="Make the output verbose (equivalent to --log DEBUG)",
    default=False,
    is_flag=True)
@click.option(
    "--profile",
    help="Print profiling information into file",
    type=click.Path(),
    default=None)
@click.option(
    "-l",
    "--lib",
    help="Choose a library name or library path (unamed library)",
    default=lambda: papis.config.getstring("default-library"))
@click.option(
    "-c",
    "--config",
    help="Configuration file to use",
    type=click.Path(exists=True),
    default=None,)
@click.option(
    "--pick-lib",
    help="Pick library to use",
    default=False,
    is_flag=True)
@click.option(
    "--cc", "--clear-cache", "clear_cache",
    help="Clear cache of the library used",
    default=False,
    is_flag=True)
@click.option(
    "-s", "--set", "set_list",
    type=(str, str),
    multiple=True,
    help="Set key value, e.g., "
         "--set info-name information.yaml  --set opentool evince",)
@click.option(
    "--color",
    type=click.Choice(["always", "auto", "no"]),
    default="auto",
    help="Prevent the output from having color")
@click.option(
    "--log",
    help="Logging level",
    type=click.Choice(["INFO", "DEBUG", "WARNING", "ERROR", "CRITICAL"]),
    default="INFO")
@click.option(
    "--logfile",
    help="File to dump the log",
    type=str,
    default=None)
def run(
        verbose: bool,
        profile: str,
        config: str,
        lib: str,
        log: str,
        logfile: Optional[str],
        pick_lib: bool,
        clear_cache: bool,
        set_list: List[Tuple[str, str]],
        color: str) -> None:

    if profile:
        profiler = cProfile.Profile()
        profiler.enable()
        atexit.register(generate_profile_writing_function(profiler, profile))

    if color == "no" or (color == "auto" and not sys.stdout.isatty()):
        # Turn off colorama (strip escape sequences from the output)
        colorama.init(strip=True)
    else:
        colorama.init()

    log_format = (colorama.Fore.YELLOW +
                  '%(levelname)s' +
                  ':' +
                  colorama.Fore.GREEN +
                  '%(name)s' +
                  colorama.Style.RESET_ALL +
                  ':' +
                  '%(message)s'
                  )
    if verbose:
        log = "DEBUG"
        log_format = '%(relativeCreated)d-'+log_format
    logging.basicConfig(level=getattr(logging, log),
                        format=log_format,
                        filename=logfile,
                        filemode='w+' if logfile is not None else 'a')
    logger = logging.getLogger('default')

    if config:
        papis.config.set_config_file(config)
        papis.config.reset_configuration()

    for pair in set_list:
        logger.debug('Setting "{0}" to "{1}"'.format(*pair))
        papis.config.set(pair[0], pair[1])

    if pick_lib:
        _picked_libs = papis.pick.pick(papis.api.get_libraries())
        if _picked_libs:
            lib = _picked_libs[0]

    papis.config.set_lib_from_name(lib)
    library = papis.config.get_lib()

    if not library.paths:
        raise Exception("Library '{0}' does not have any existing folders"
                        " attached to it, please define and create the paths"
                        .format(lib))

    for path in library.paths:
        # Now the library should be set, let us check if there is a
        # local configuration file there, and if there is one, then
        # merge its contents
        local_config_file = os.path.expanduser(
            os.path.join(path,
                         papis.config.getstring("local-config-file")))
        papis.config.merge_configuration_from_path(
            local_config_file,
            papis.config.get_configuration())

    if clear_cache:
        papis.database.get().clear()



import re
import json
from typing import Dict, Optional, Any, List

import papis.downloaders
import papis.document

_K = papis.document.KeyConversionPair


def _page_to_pages(data: List[Dict[str, str]]) -> str:
    if len(data) == 0:
        raise RuntimeError("No data to turn to pages")
    x = data[0]
    if not {"first-page", "last-page"} & x.keys():
        raise RuntimeError("first-page and last-page not found")
    return "{0}--{1}".format(x["first-page"], x["last-page"])


article_keyconv = [
    _K("doi", [papis.document.EmptyKeyConversion]),
    _K("pii", [papis.document.EmptyKeyConversion]),
    _K("language", [papis.document.EmptyKeyConversion]),
    _K("subtitle", [papis.document.EmptyKeyConversion]),
    _K("issn", [papis.document.EmptyKeyConversion]),
    _K("srctitle", [{"key": "journal", "action": None}]),
    _K("vol-first", [{"key": "volume", "action": None}]),
    _K("cover-date-years", [{"key": "year", "action": lambda x: x[0]}]),
    _K("cover-date-start", [{"key": "date", "action": None}]),
    _K("document-type", [{"key": "type", "action": None}]),
    _K("titleString", [{"key": "title", "action": None}]),
    _K("dates", [
        {"key": "accepted-date", "action": lambda x: x["Accepted"]},
        {"key": "publication-date", "action": lambda x: x["Publication date"]}
    ]),
    _K("pages", [{
        "action": _page_to_pages,
        "key": None
    }]),
]  # List[papis.document.KeyConversionPair]
script_keyconv = [
    _K("article", [{
        "key": "_article_data",
        "action": lambda x:
            papis.document.keyconversion_to_data(article_keyconv, x)
    }]),
]  # List[papis.document.KeyConversionPair]


class Downloader(papis.downloaders.Downloader):

    def __init__(self, url: str):
        papis.downloaders.Downloader.__init__(
            self, url, name="sciencedirect")
        self.expected_document_extension = 'pdf'

    @classmethod
    def match(cls, url: str) -> Optional[papis.downloaders.Downloader]:
        if re.match(r".*\.sciencedirect\.com.*", url):
            return Downloader(url)
        else:
            return None

    def get_data(self) -> Dict[str, Any]:
        global script_keyconv
        data = dict()  # type: Dict[str, Any]
        soup = self._get_soup()
        scripts = soup.find_all(name="script", attrs={'data-iso-key': '_0'})
        if scripts:
            rawdata = json.loads(scripts[0].text)
            self.logger.debug(
                "found {0} scripts data-iso-key".format(len(scripts)))
            converted_data = papis.document.keyconversion_to_data(
                script_keyconv, rawdata)
            data.update(converted_data["_article_data"])
            # TODO: parse abstract and author in a typed checked and meaningful
            #       way
            # data.update(converted_data["_abstract_data"])
            # data.update(converted_data["_author_data"])
            data['author'] = papis.document.author_list_to_author(data)
        return data



"""
The export command is useful to work with other programs such as bibtex.

Some examples of its usage are:

- Export one of the documents matching the author with einstein to bibtex:

.. code::

    papis export --format bibtex 'author : einstein'

or export all of them

.. code::

    papis export --format bibtex --all 'author : einstein'

- Export all documents to bibtex and save them into a ``lib.bib`` file

.. code::

    papis export --all --format bibtex --out lib.bib

- Export a folder of one of the documents matching the word ``krebs``
  into a folder named, ``interesting-document``

.. code::

    papis export --folder --out interesting-document krebs

  this will create the folder ``interesting-document`` containing the
  ``info.yaml`` file, the linked documents and a ``bibtex`` file for
  sharing with other people.

.. note::

    Every document exported also comes with the key `_papis_local_folder`
    associated that points to the full local folder path where the document
    is stored in the file system. This is done for the convenience of third
    party apps.


Cli
^^^
.. click:: papis.commands.export:cli
    :prog: papis export
"""
import papis
import os
import shutil
import papis.utils
import papis.document
import click
import papis.cli
import papis.api
import papis.database
import papis.strings
import logging
import papis.plugin
from typing import List, Optional

logger = logging.getLogger('cli:export')


def available_formats() -> List[str]:
    return papis.plugin.get_available_entrypoints(_extension_name())


def _extension_name() -> str:
    return "papis.exporter"


def run(documents: List[papis.document.Document], to_format: str,) -> str:
    """
    Exports several documents into something else.

    :param document: A ist of papis document
    :type  document: [papis.document.Document]
    :param to_format: what format to use
    :type  to_format: str
    """
    ret_string = (
        papis.plugin.get_extension_manager(_extension_name())[to_format]
        .plugin(document for document in documents))
    return str(ret_string)


@click.command("export")
@click.help_option('--help', '-h')
@papis.cli.query_option()
@papis.cli.doc_folder_option()
@papis.cli.sort_option()
@papis.cli.all_option()
@click.option(
    "--folder",
    help="Export document folder to share",
    default=False,
    is_flag=True)
@click.option(
    "-o",
    "--out",
    help="Outfile or outdir",
    default=None)
@click.option(
    "-f",
    "--format", "fmt",
    help="Format for the document",
    type=click.Choice(available_formats()),
    default="bibtex",)
def cli(query: str,
        doc_folder: str,
        sort_field: Optional[str],
        sort_reverse: bool,
        folder: str,
        out: str,
        fmt: str,
        _all: bool) -> None:
    """Export a document from a given library"""

    if doc_folder:
        documents = [papis.document.from_folder(doc_folder)]
    else:
        documents = papis.database.get().query(query)

    if fmt and folder:
        logger.warning("Only --folder flag will be considered")

    if not documents:
        logger.warning(papis.strings.no_documents_retrieved_message)
        return

    if not _all:
        documents = [d for d in papis.pick.pick_doc(documents)]
        if not documents:
            return

    if sort_field:
        documents = papis.document.sort(documents, sort_field, sort_reverse)

    # Get the local folder of the document so that third-party apps
    # can actually go to the folder without checking with papis
    for d in documents:
        d["_papis_local_folder"] = d.get_main_folder()

    ret_string = run(documents, to_format=fmt)

    if ret_string is not None and not folder:
        if out is not None:
            logger.info("Dumping to {0}".format(out))
            with open(out, 'a+') as fd:
                fd.write(ret_string)
        else:
            logger.info("Dumping to stdout")
            print(ret_string)
        return

    for document in documents:
        if folder:
            _doc_folder = document.get_main_folder()
            _doc_folder_name = document.get_main_folder_name()
            outdir = out or _doc_folder_name
            if not _doc_folder or not _doc_folder_name or not outdir:
                raise Exception(papis.strings.no_folder_attached_to_document)
            if not len(documents) == 1:
                outdir = os.path.join(out, _doc_folder_name)
            logger.info("Exporting doc {0} to {1}".format(
                papis.document.describe(document), outdir
            ))
            shutil.copytree(_doc_folder, outdir)


@click.command('export')
@click.pass_context
@click.help_option('--help', '-h')
@click.option(
    "-f", "--format", "fmt",
    help="Format for the document",
    type=click.Choice(available_formats()),
    default="bibtex",)
@click.option(
    "-o",
    "--out",
    help="Outfile to write information to",
    type=click.Path(),
    default=None,)
def explorer(ctx: click.Context, fmt: str, out: str) -> None:
    """
    Export retrieved documents into various formats for later use

    Examples of its usage are

    papis explore crossref -m 200 -a 'Schrodinger' export --yaml lib.yaml

    """
    logger = logging.getLogger('explore:yaml')
    docs = ctx.obj['documents']

    outstring = run(docs, to_format=fmt)
    if out is not None:
        with open(out, 'a+') as fd:
            logger.info(
                "Writing {} documents' in {} into {}".format(
                    len(docs), fmt, out))
            fd.write(outstring)
    else:
        print(outstring)



import re
import papis.downloaders.fallback

from typing import Optional


class Downloader(papis.downloaders.fallback.Downloader):

    def __init__(self, url: str):
        papis.downloaders.fallback.Downloader.__init__(
            self,
            uri=url,
            name="aps")
        self.expected_document_extension = 'pdf'
        self.priority = 10

    @classmethod
    def match(
            cls, url: str) -> Optional[papis.downloaders.fallback.Downloader]:
        return Downloader(url) if re.match(r".*aps.org.*", url) else None

    def get_bibtex_url(self) -> Optional[str]:
        burl = "{}?{}".format(
            re.sub(r'/abstract', r'/export', self.uri),
            "type=bibtex&download=true")
        self.logger.debug("[bibtex url] = %s" % burl)
        return burl



import logging
import os
import functools
from typing import Callable, TypeVar, Generic, Sequence, Type
from abc import ABC, abstractmethod

import papis.config
import papis.document
import papis.plugin

LOGGER = logging.getLogger("pick")
T = TypeVar("T")
Option = TypeVar("Option")


class Picker(ABC, Generic[T]):

    @abstractmethod
    def __call__(
            self,
            items: Sequence[T],
            header_filter: Callable[[T], str],
            match_filter: Callable[[T], str],
            default_index: int = 0
            ) -> Sequence[T]:
        ...


def _extension_name() -> str:
    return "papis.picker"


def get_picker(name: str) -> Type[Picker[Option]]:
    """Get the picker named 'name' declared as a plugin"""
    picker = papis.plugin.get_extension_manager(
        _extension_name())[name].plugin  # type: Type[Picker[Option]]
    return picker


def pick(
        options: Sequence[Option],
        default_index: int = 0,
        header_filter: Callable[[Option], str] = str,
        match_filter: Callable[[Option], str] = str) -> Sequence[Option]:

    name = papis.config.getstring("picktool")
    try:
        picker = get_picker(name)  # type: Type[Picker[Option]]
    except KeyError:
        LOGGER.error("Invalid picker (%s)", name)
        LOGGER.error(
            "Registered pickers are: %s",
            papis.plugin.get_available_entrypoints(_extension_name()))
        return []
    else:
        return picker()(options,
                        header_filter,
                        match_filter,
                        default_index)


def pick_doc(
        documents: Sequence[papis.document.Document]
     ) -> Sequence[papis.document.Document]:
    """Pick a document from documents with the correct formatting

    :documents: List of documents
    :returns: Document

    """
    header_format_path = papis.config.get('header-format-file')
    if header_format_path is not None:
        with open(os.path.expanduser(header_format_path)) as _fd:
            header_format = _fd.read()
    else:
        header_format = papis.config.getstring("header-format")
    match_format = papis.config.getstring("match-format")
    header_filter = functools.partial(papis.format.format, header_format)
    match_filter = functools.partial(papis.format.format, match_format)
    return pick(documents,
                header_filter=header_filter,
                match_filter=match_filter)



import re
import papis.downloaders.base
from typing import Optional


class Downloader(papis.downloaders.Downloader):

    def __init__(self, url: str):
        papis.downloaders.Downloader.__init__(self, url, name="get")
        self.priority = 0

    @classmethod
    def match(cls, url: str) -> Optional[papis.downloaders.Downloader]:
        """
        >>> Downloader.match('http://wha2341!@#!@$%!@#file.pdf') is False
        False
        >>> Downloader.match('https://whateverpt?is?therefile.epub') is False
        False
        >>> not Downloader.match('http://whatever?path?is?therefile')
        True
        """
        endings = "pdf|djvu|epub|mobi|jpg|png|md"
        m = re.match(r"^http.*\.(%s)$" % endings, url, re.IGNORECASE)
        if m:
            d = Downloader(url)
            extension = m.group(1)
            d.logger.info(
                'Expecting a document of type "{0}"'.format(extension))
            d.expected_document_extension = extension
            return d
        else:
            return None

    def get_document_url(self) -> Optional[str]:
        return self.uri



"""

Cli
^^^
.. click:: papis.commands.rename:cli
    :prog: papis rename
"""
import os
import subprocess
import logging
import click
import papis.cli
import papis.database
import papis.strings
import papis.git
import papis.pick
import papis.document
import papis.tui.utils

from typing import Optional


def run(document: papis.document.Document,
        new_name: str, git: bool = False) -> None:
    db = papis.database.get()
    logger = logging.getLogger('rename:run')
    folder = document.get_main_folder()

    if not folder:
        raise Exception(papis.strings.no_folder_attached_to_document)

    subfolder = os.path.dirname(folder)

    new_folder_path = os.path.join(subfolder, new_name)

    if os.path.exists(new_folder_path):
        logger.warning("Path %s already exists" % new_folder_path)
        return

    cmd = ['git', '-C', folder] if git else []
    cmd += ['mv', folder, new_folder_path]

    logger.debug(cmd)
    subprocess.call(cmd)

    if git:
        papis.git.commit(
            new_folder_path,
            "Rename from {} to '{}'".format(folder, new_name))

    db.delete(document)
    logger.debug("New document folder: {}".format(new_folder_path))
    document.set_folder(new_folder_path)
    db.add(document)


@click.command("rename")
@click.help_option('--help', '-h')
@papis.cli.query_option()
@papis.cli.git_option()
@papis.cli.sort_option()
@papis.cli.doc_folder_option()
def cli(query: str,
        git: bool,
        sort_field: Optional[str],
        doc_folder: str,
        sort_reverse: bool) -> None:
    """Rename entry"""

    if doc_folder:
        documents = [papis.document.from_folder(doc_folder)]
    else:
        documents = papis.database.get().query(query)

    if sort_field:
        documents = papis.document.sort(documents, sort_field, sort_reverse)

    logger = logging.getLogger('cli:rename')

    if not documents:
        logger.warning(papis.strings.no_documents_retrieved_message)
    docs = papis.pick.pick_doc(documents)
    if not docs:
        return

    document = docs[0]

    new_name = papis.tui.utils.prompt(
        "Enter new folder name:\n"
        ">",
        default=document.get_main_folder_name() or ''
    )
    run(document, new_name, git=git)



import papis.document
import logging
import isbnlib
import click
# See https://github.com/xlcnd/isbnlib for details
from typing import Dict, Any, List

logger = logging.getLogger('papis:isbnlib')


def get_data(query: str = "",
             service: str = 'openl') -> List[Dict[str, Any]]:
    global logger
    results = []  # type: List[Dict[str, Any]]
    logger.debug('Trying to retrieve isbn')
    isbn = isbnlib.isbn_from_words(query)
    data = isbnlib.meta(isbn, service=service)
    if data is None:
        return results
    else:
        logger.debug('Trying to retrieve isbn')
        assert(isinstance(data, dict))
        results.append(data_to_papis(data))
        return results


def data_to_papis(data: Dict[str, Any]) -> Dict[str, Any]:
    """Convert data from isbnlib into papis formated data

    :param data: Dictionary with data
    :type  data: dict
    :returns: Dictionary with papis keynames

    """
    data = {k.lower(): data[k] for k in data}
    return data


@click.command('isbn')
@click.pass_context
@click.help_option('--help', '-h')
@click.option('--query', '-q', default=None)
@click.option('--service', '-s',
              default='goob',
              type=click.Choice(['wcat', 'goob', 'openl']))
def explorer(ctx: click.core.Context, query: str, service: str) -> None:
    """
    Look for documents using isbnlib

    Examples of its usage are

    papis explore isbn -q 'Albert einstein' pick cmd 'firefox {doc[url]}'

    """
    logger = logging.getLogger('explore:isbn')
    logger.info('Looking up...')
    data = get_data(query=query, service=service)
    docs = [papis.document.from_data(data=d) for d in data]
    logger.info('{} documents found'.format(len(docs)))
    ctx.obj['documents'] += docs



import click
import json
import logging
from typing import List

import papis.document


def exporter(documents: List[papis.document.Document]) -> str:
    return json.dumps([papis.document.to_dict(doc) for doc in documents])


@click.command('json')
@click.pass_context
@click.argument('jsonfile', type=click.Path(exists=True))
@click.help_option('--help', '-h')
def explorer(ctx: click.Context, jsonfile: str) -> None:
    """
    Import documents from a json file

    Examples of its usage are

    papis explore json lib.json pick

    """
    logger = logging.getLogger('explore:json')
    logger.info('Reading in json file {}'.format(jsonfile))
    docs = [papis.document.from_data(d) for d in json.load(open(jsonfile))]
    ctx.obj['documents'] += docs
    logger.info('{} documents found'.format(len(docs)))
